"""
The linear harmonic objective function.

The bit at index `i` for `i` in `0..n-1` has weight `i+1`. This is the penalty
that is incurred if the bit is set to `False`.
The best objective value, 0, is hence obtained if all bits are `True`.
The worst objective value, i.e., `n * (n + 1) // 2`, is obtained if all bits
are `False`.

1. Carola Doerr, Furong Ye, Naama Horesh, Hao Wang, Ofer M. Shir, and Thomas
   Bäck. Benchmarking Discrete Optimization Heuristics with IOHprofiler.
   Applied Soft Computing Journal. 88:106027. 2020.
   doi: https://doi.org/10.1016/j.asoc.2019.106027
2. Stefan Droste, Thomas Jansen, and Ingo Wegener. On the Analysis of the
   (1+1) Evolutionary Algorithm. *Theoretical Computer Science.*
   276(1-2):51-81. April 2002.
   doi: https://doi.org/10.1016/S0304-3975(01)00182-7
3. Thomas Weise, Zhize Wu, Xinlu Li, Yan Chen, and Jörg Lässig. Frequency
   Fitness Assignment: Optimization without Bias for Good Solutions can be
   Efficient. *IEEE Transactions on Evolutionary Computation (TEVC)*.
   27(4):980-992. August 2023.
   doi: https://doi.org/10.1109/TEVC.2022.3191698

This is code is part of the research work of Mr. Jiazheng ZENG (曾嘉政),
a Master's student at the Institute of Applied Optimization
(应用优化研究所) of the School of Artificial
Intelligence and Big Data (人工智能与大数据学院) at
Hefei University (合肥大学) in
Hefei, Anhui, China (中国安徽省合肥市) under the supervision of
Prof. Dr. Thomas Weise (汤卫思教授).
"""

from typing import Callable, Iterator, cast

import numba  # type: ignore
import numpy as np

from moptipy.examples.bitstrings.bitstring_problem import BitStringProblem


@numba.njit(nogil=True, cache=True)
def linear_harmonic(x: np.ndarray) -> float:
    """
    Evaluate the linear function with harmonic weights.

    :param  x: np array representing the bit string
    :return: the objective value

    >>> linear_harmonic(np.array([True, True, True]))
    0
    >>> linear_harmonic(np.array([False, True, True]))
    1
    >>> linear_harmonic(np.array([True, False, True]))
    2
    >>> linear_harmonic(np.array([True, True, False]))
    3
    >>> linear_harmonic(np.array([False, False, True]))
    3
    >>> linear_harmonic(np.array([False, True, False]))
    4
    >>> linear_harmonic(np.array([True, False, False]))
    5
    >>> linear_harmonic(np.array([False, False, False]))
    6
    >>> (3 * (3 + 1)) // 2
    6

    >>> linear_harmonic(np.array([True, True, True, True]))
    0
    >>> linear_harmonic(np.array([False, True, True, True]))
    1
    >>> linear_harmonic(np.array([True, False, True, True]))
    2
    >>> linear_harmonic(np.array([True, True, False, True]))
    3
    >>> linear_harmonic(np.array([True, True, True, False]))
    4
    >>> linear_harmonic(np.array([False, False, True, True]))
    3
    >>> linear_harmonic(np.array([False, True, False, True]))
    4
    >>> linear_harmonic(np.array([False, True, True, False]))
    5
    >>> linear_harmonic(np.array([True, False, False, True]))
    5
    >>> linear_harmonic(np.array([True, False, True, False]))
    6
    >>> linear_harmonic(np.array([True, True, False, False]))
    7
    >>> linear_harmonic(np.array([False, False, False, True]))
    6
    >>> linear_harmonic(np.array([False, False, True, False]))
    7
    >>> linear_harmonic(np.array([False, True, False, False]))
    8
    >>> linear_harmonic(np.array([True, False, False, False]))
    9
    >>> linear_harmonic(np.array([False, False, False, False]))
    10
    >>> 4 * (4 + 1) // 2
    10

    # n = 1 and 0 true bits
    >>> linear_harmonic(np.array([0]))
    1

    # n = 1 and 1 true bit
    >>> linear_harmonic(np.array([1]))
    0

    # n = 2 and 0 true bits
    >>> linear_harmonic(np.array([0, 0]))
    3

    # n = 2 and 1 true bit
    >>> linear_harmonic(np.array([1, 0]))
    2

    # n = 2 and 1 true bit
    >>> linear_harmonic(np.array([1, 0]))
    2

    # n = 2 and 1 true bit
    >>> linear_harmonic(np.array([1, 0]))
    2

    # n = 2 and 2 true bits
    >>> linear_harmonic(np.array([1, 1]))
    0

    # n = 3 and 0 true bits
    >>> linear_harmonic(np.array([0, 0, 0]))
    6

    # n = 3 and 1 true bit
    >>> linear_harmonic(np.array([0, 1, 0]))
    4

    # n = 3 and 1 true bit
    >>> linear_harmonic(np.array([0, 0, 1]))
    3

    # n = 3 and 1 true bit
    >>> linear_harmonic(np.array([1, 0, 0]))
    5

    # n = 3 and 2 true bits
    >>> linear_harmonic(np.array([0, 1, 1]))
    1

    # n = 3 and 2 true bits
    >>> linear_harmonic(np.array([1, 0, 1]))
    2

    # n = 3 and 2 true bits
    >>> linear_harmonic(np.array([1, 1, 0]))
    3

    # n = 3 and 3 true bits
    >>> linear_harmonic(np.array([1, 1, 1]))
    0

    # n = 4 and 0 true bits
    >>> linear_harmonic(np.array([0, 0, 0, 0]))
    10

    # n = 4 and 1 true bit
    >>> linear_harmonic(np.array([1, 0, 0, 0]))
    9

    # n = 4 and 1 true bit
    >>> linear_harmonic(np.array([1, 0, 0, 0]))
    9

    # n = 4 and 1 true bit
    >>> linear_harmonic(np.array([1, 0, 0, 0]))
    9

    # n = 4 and 2 true bits
    >>> linear_harmonic(np.array([0, 1, 1, 0]))
    5

    # n = 4 and 2 true bits
    >>> linear_harmonic(np.array([0, 1, 0, 1]))
    4

    # n = 4 and 2 true bits
    >>> linear_harmonic(np.array([1, 0, 0, 1]))
    5

    # n = 4 and 3 true bits
    >>> linear_harmonic(np.array([1, 1, 0, 1]))
    3

    # n = 4 and 3 true bits
    >>> linear_harmonic(np.array([0, 1, 1, 1]))
    1

    # n = 4 and 3 true bits
    >>> linear_harmonic(np.array([1, 0, 1, 1]))
    2

    # n = 4 and 4 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1]))
    0

    # n = 5 and 0 true bits
    >>> linear_harmonic(np.array([0, 0, 0, 0, 0]))
    15

    # n = 5 and 1 true bit
    >>> linear_harmonic(np.array([0, 1, 0, 0, 0]))
    13

    # n = 5 and 1 true bit
    >>> linear_harmonic(np.array([1, 0, 0, 0, 0]))
    14

    # n = 5 and 2 true bits
    >>> linear_harmonic(np.array([0, 0, 1, 0, 1]))
    7

    # n = 5 and 2 true bits
    >>> linear_harmonic(np.array([0, 1, 0, 1, 0]))
    9

    # n = 5 and 2 true bits
    >>> linear_harmonic(np.array([1, 0, 0, 1, 0]))
    10

    # n = 5 and 3 true bits
    >>> linear_harmonic(np.array([0, 0, 1, 1, 1]))
    3

    # n = 5 and 3 true bits
    >>> linear_harmonic(np.array([0, 0, 1, 1, 1]))
    3

    # n = 5 and 3 true bits
    >>> linear_harmonic(np.array([1, 0, 1, 0, 1]))
    6

    # n = 5 and 4 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 0]))
    5

    # n = 5 and 4 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 0]))
    5

    # n = 5 and 4 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 0]))
    5

    # n = 5 and 5 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 1]))
    0

    # n = 6 and 0 true bits
    >>> linear_harmonic(np.array([0, 0, 0, 0, 0, 0]))
    21

    # n = 6 and 1 true bit
    >>> linear_harmonic(np.array([0, 0, 0, 0, 1, 0]))
    16

    # n = 6 and 1 true bit
    >>> linear_harmonic(np.array([0, 0, 1, 0, 0, 0]))
    18

    # n = 6 and 1 true bit
    >>> linear_harmonic(np.array([0, 0, 0, 0, 0, 1]))
    15

    # n = 6 and 2 true bits
    >>> linear_harmonic(np.array([1, 0, 1, 0, 0, 0]))
    17

    # n = 6 and 2 true bits
    >>> linear_harmonic(np.array([0, 0, 1, 0, 1, 0]))
    13

    # n = 6 and 2 true bits
    >>> linear_harmonic(np.array([0, 0, 0, 0, 1, 1]))
    10

    # n = 6 and 3 true bits
    >>> linear_harmonic(np.array([0, 1, 0, 0, 1, 1]))
    8

    # n = 6 and 3 true bits
    >>> linear_harmonic(np.array([1, 1, 0, 0, 0, 1]))
    12

    # n = 6 and 3 true bits
    >>> linear_harmonic(np.array([1, 1, 0, 0, 1, 0]))
    13

    # n = 6 and 4 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 0, 0]))
    11

    # n = 6 and 4 true bits
    >>> linear_harmonic(np.array([1, 1, 0, 1, 1, 0]))
    9

    # n = 6 and 4 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 0, 0, 1]))
    9

    # n = 6 and 5 true bits
    >>> linear_harmonic(np.array([1, 1, 0, 1, 1, 1]))
    3

    # n = 6 and 5 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 1, 0]))
    6

    # n = 6 and 5 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 0, 1]))
    5

    # n = 6 and 6 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 1, 1]))
    0

    # n = 7 and 0 true bits
    >>> linear_harmonic(np.array([0, 0, 0, 0, 0, 0, 0]))
    28

    # n = 7 and 1 true bit
    >>> linear_harmonic(np.array([0, 0, 1, 0, 0, 0, 0]))
    25

    # n = 7 and 1 true bit
    >>> linear_harmonic(np.array([0, 0, 0, 1, 0, 0, 0]))
    24

    # n = 7 and 1 true bit
    >>> linear_harmonic(np.array([1, 0, 0, 0, 0, 0, 0]))
    27

    # n = 7 and 2 true bits
    >>> linear_harmonic(np.array([0, 1, 0, 0, 0, 1, 0]))
    20

    # n = 7 and 2 true bits
    >>> linear_harmonic(np.array([0, 0, 0, 0, 1, 0, 1]))
    16

    # n = 7 and 2 true bits
    >>> linear_harmonic(np.array([1, 0, 0, 0, 0, 1, 0]))
    21

    # n = 7 and 3 true bits
    >>> linear_harmonic(np.array([0, 1, 0, 0, 0, 1, 1]))
    13

    # n = 7 and 3 true bits
    >>> linear_harmonic(np.array([1, 0, 0, 1, 0, 0, 1]))
    16

    # n = 7 and 3 true bits
    >>> linear_harmonic(np.array([0, 1, 0, 0, 0, 1, 1]))
    13

    # n = 7 and 4 true bits
    >>> linear_harmonic(np.array([1, 1, 0, 1, 0, 0, 1]))
    14

    # n = 7 and 4 true bits
    >>> linear_harmonic(np.array([1, 0, 1, 0, 1, 0, 1]))
    12

    # n = 7 and 4 true bits
    >>> linear_harmonic(np.array([0, 0, 1, 1, 1, 1, 0]))
    10

    # n = 7 and 5 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 0, 0, 1, 1]))
    9

    # n = 7 and 5 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 0, 1, 0]))
    12

    # n = 7 and 5 true bits
    >>> linear_harmonic(np.array([1, 1, 0, 1, 0, 1, 1]))
    8

    # n = 7 and 6 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 1, 1, 0]))
    7

    # n = 7 and 6 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 1, 1, 0]))
    7

    # n = 7 and 6 true bits
    >>> linear_harmonic(np.array([1, 1, 0, 1, 1, 1, 1]))
    3

    # n = 7 and 7 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 1, 1, 1]))
    0

    # n = 8 and 0 true bits
    >>> linear_harmonic(np.array([0, 0, 0, 0, 0, 0, 0, 0]))
    36

    # n = 8 and 1 true bit
    >>> linear_harmonic(np.array([0, 1, 0, 0, 0, 0, 0, 0]))
    34

    # n = 8 and 1 true bit
    >>> linear_harmonic(np.array([0, 0, 0, 0, 1, 0, 0, 0]))
    31

    # n = 8 and 1 true bit
    >>> linear_harmonic(np.array([0, 1, 0, 0, 0, 0, 0, 0]))
    34

    # n = 8 and 2 true bits
    >>> linear_harmonic(np.array([0, 0, 1, 1, 0, 0, 0, 0]))
    29

    # n = 8 and 2 true bits
    >>> linear_harmonic(np.array([0, 0, 1, 0, 0, 0, 1, 0]))
    26

    # n = 8 and 2 true bits
    >>> linear_harmonic(np.array([0, 0, 0, 0, 0, 1, 0, 1]))
    22

    # n = 8 and 3 true bits
    >>> linear_harmonic(np.array([1, 1, 0, 0, 0, 0, 0, 1]))
    25

    # n = 8 and 3 true bits
    >>> linear_harmonic(np.array([1, 1, 0, 0, 0, 0, 0, 1]))
    25

    # n = 8 and 3 true bits
    >>> linear_harmonic(np.array([0, 0, 0, 0, 1, 1, 0, 1]))
    17

    # n = 8 and 4 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 0, 0, 0, 0, 1]))
    22

    # n = 8 and 4 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 0, 0, 0, 0, 1]))
    22

    # n = 8 and 4 true bits
    >>> linear_harmonic(np.array([1, 0, 1, 0, 0, 1, 0, 1]))
    18

    # n = 8 and 5 true bits
    >>> linear_harmonic(np.array([0, 1, 1, 1, 0, 1, 0, 1]))
    13

    # n = 8 and 5 true bits
    >>> linear_harmonic(np.array([0, 0, 1, 0, 1, 1, 1, 1]))
    7

    # n = 8 and 5 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 0, 0, 1, 0]))
    19

    # n = 8 and 6 true bits
    >>> linear_harmonic(np.array([0, 1, 1, 1, 1, 0, 1, 1]))
    7

    # n = 8 and 6 true bits
    >>> linear_harmonic(np.array([0, 1, 1, 1, 1, 1, 1, 0]))
    9

    # n = 8 and 6 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 0, 1, 1, 0]))
    13

    # n = 8 and 7 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 1, 1, 0, 1]))
    7

    # n = 8 and 7 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 0, 1, 1, 1, 1]))
    4

    # n = 8 and 7 true bits
    >>> linear_harmonic(np.array([1, 0, 1, 1, 1, 1, 1, 1]))
    2

    # n = 8 and 8 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 1, 1, 1, 1]))
    0

    # n = 9 and 0 true bits
    >>> linear_harmonic(np.array([0, 0, 0, 0, 0, 0, 0, 0, 0]))
    45

    # n = 9 and 1 true bit
    >>> linear_harmonic(np.array([0, 1, 0, 0, 0, 0, 0, 0, 0]))
    43

    # n = 9 and 1 true bit
    >>> linear_harmonic(np.array([0, 0, 0, 0, 0, 0, 0, 1, 0]))
    37

    # n = 9 and 1 true bit
    >>> linear_harmonic(np.array([0, 0, 0, 0, 0, 1, 0, 0, 0]))
    39

    # n = 9 and 2 true bits
    >>> linear_harmonic(np.array([0, 0, 0, 1, 0, 1, 0, 0, 0]))
    35

    # n = 9 and 2 true bits
    >>> linear_harmonic(np.array([1, 0, 0, 1, 0, 0, 0, 0, 0]))
    40

    # n = 9 and 2 true bits
    >>> linear_harmonic(np.array([1, 1, 0, 0, 0, 0, 0, 0, 0]))
    42

    # n = 9 and 3 true bits
    >>> linear_harmonic(np.array([0, 0, 0, 0, 1, 0, 0, 1, 1]))
    23

    # n = 9 and 3 true bits
    >>> linear_harmonic(np.array([0, 0, 1, 1, 1, 0, 0, 0, 0]))
    33

    # n = 9 and 3 true bits
    >>> linear_harmonic(np.array([0, 1, 1, 0, 0, 0, 1, 0, 0]))
    33

    # n = 9 and 4 true bits
    >>> linear_harmonic(np.array([1, 0, 1, 0, 0, 1, 0, 1, 0]))
    27

    # n = 9 and 4 true bits
    >>> linear_harmonic(np.array([0, 0, 1, 1, 0, 1, 0, 0, 1]))
    23

    # n = 9 and 4 true bits
    >>> linear_harmonic(np.array([0, 1, 0, 1, 0, 1, 0, 0, 1]))
    24

    # n = 9 and 5 true bits
    >>> linear_harmonic(np.array([1, 0, 0, 1, 0, 1, 1, 0, 1]))
    18

    # n = 9 and 5 true bits
    >>> linear_harmonic(np.array([1, 0, 0, 1, 1, 0, 0, 1, 1]))
    18

    # n = 9 and 5 true bits
    >>> linear_harmonic(np.array([0, 1, 1, 1, 0, 1, 0, 0, 1]))
    21

    # n = 9 and 6 true bits
    >>> linear_harmonic(np.array([1, 1, 0, 0, 1, 1, 1, 1, 0]))
    16

    # n = 9 and 6 true bits
    >>> linear_harmonic(np.array([1, 1, 0, 1, 0, 1, 1, 0, 1]))
    16

    # n = 9 and 6 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 0, 1, 0, 1, 1, 0]))
    19

    # n = 9 and 7 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 1, 1, 0, 1, 0]))
    16

    # n = 9 and 7 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 1, 1, 1, 0, 0]))
    17

    # n = 9 and 7 true bits
    >>> linear_harmonic(np.array([0, 1, 0, 1, 1, 1, 1, 1, 1]))
    4

    # n = 9 and 8 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 1, 0, 1, 1, 1]))
    6

    # n = 9 and 8 true bits
    >>> linear_harmonic(np.array([0, 1, 1, 1, 1, 1, 1, 1, 1]))
    1

    # n = 9 and 8 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 1, 1, 0, 1, 1]))
    7

    # n = 9 and 9 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 1, 1, 1, 1, 1]))
    0

    # n = 10 and 0 true bits
    >>> linear_harmonic(np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0]))
    55

    # n = 10 and 1 true bit
    >>> linear_harmonic(np.array([0, 0, 0, 0, 0, 0, 0, 0, 1, 0]))
    46

    # n = 10 and 1 true bit
    >>> linear_harmonic(np.array([0, 0, 0, 0, 1, 0, 0, 0, 0, 0]))
    50

    # n = 10 and 1 true bit
    >>> linear_harmonic(np.array([0, 1, 0, 0, 0, 0, 0, 0, 0, 0]))
    53

    # n = 10 and 2 true bits
    >>> linear_harmonic(np.array([0, 0, 1, 0, 0, 1, 0, 0, 0, 0]))
    46

    # n = 10 and 2 true bits
    >>> linear_harmonic(np.array([0, 0, 1, 0, 1, 0, 0, 0, 0, 0]))
    47

    # n = 10 and 2 true bits
    >>> linear_harmonic(np.array([0, 0, 0, 0, 0, 0, 0, 0, 1, 1]))
    36

    # n = 10 and 3 true bits
    >>> linear_harmonic(np.array([1, 0, 0, 0, 0, 0, 1, 0, 0, 1]))
    37

    # n = 10 and 3 true bits
    >>> linear_harmonic(np.array([0, 0, 0, 1, 0, 0, 0, 1, 1, 0]))
    34

    # n = 10 and 3 true bits
    >>> linear_harmonic(np.array([0, 0, 1, 1, 0, 0, 0, 0, 0, 1]))
    38

    # n = 10 and 4 true bits
    >>> linear_harmonic(np.array([0, 1, 1, 0, 0, 0, 0, 1, 1, 0]))
    33

    # n = 10 and 4 true bits
    >>> linear_harmonic(np.array([0, 1, 0, 0, 0, 1, 1, 0, 1, 0]))
    31

    # n = 10 and 4 true bits
    >>> linear_harmonic(np.array([0, 1, 1, 0, 0, 1, 0, 0, 0, 1]))
    34

    # n = 10 and 5 true bits
    >>> linear_harmonic(np.array([1, 0, 0, 1, 0, 0, 1, 1, 1, 0]))
    26

    # n = 10 and 5 true bits
    >>> linear_harmonic(np.array([0, 0, 1, 0, 1, 1, 1, 0, 1, 0]))
    25

    # n = 10 and 5 true bits
    >>> linear_harmonic(np.array([0, 0, 0, 1, 1, 0, 0, 1, 1, 1]))
    19

    # n = 10 and 6 true bits
    >>> linear_harmonic(np.array([1, 1, 0, 0, 1, 1, 1, 0, 0, 1]))
    24

    # n = 10 and 6 true bits
    >>> linear_harmonic(np.array([0, 1, 1, 1, 1, 0, 1, 0, 0, 1]))
    24

    # n = 10 and 6 true bits
    >>> linear_harmonic(np.array([0, 0, 1, 0, 1, 1, 1, 1, 0, 1]))
    16

    # n = 10 and 7 true bits
    >>> linear_harmonic(np.array([1, 0, 1, 1, 1, 0, 1, 1, 1, 0]))
    18

    # n = 10 and 7 true bits
    >>> linear_harmonic(np.array([1, 0, 1, 0, 1, 1, 0, 1, 1, 1]))
    13

    # n = 10 and 7 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 1, 1, 0, 1, 0, 0]))
    26

    # n = 10 and 8 true bits
    >>> linear_harmonic(np.array([1, 1, 0, 1, 1, 1, 1, 1, 1, 0]))
    13

    # n = 10 and 8 true bits
    >>> linear_harmonic(np.array([1, 1, 0, 1, 1, 1, 1, 0, 1, 1]))
    11

    # n = 10 and 8 true bits
    >>> linear_harmonic(np.array([1, 0, 1, 1, 1, 0, 1, 1, 1, 1]))
    8

    # n = 10 and 9 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 1, 1, 1, 1, 0, 1]))
    9

    # n = 10 and 9 true bits
    >>> linear_harmonic(np.array([1, 1, 0, 1, 1, 1, 1, 1, 1, 1]))
    3

    # n = 10 and 9 true bits
    >>> linear_harmonic(np.array([1, 1, 0, 1, 1, 1, 1, 1, 1, 1]))
    3

    # n = 10 and 10 true bits
    >>> linear_harmonic(np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1]))
    0
    """
    result: int = 0  # The sum of the penalty weights
    weight: int = 1  # The current penalty weight.

    for xx in x:  # Iterate over all the bits in the array.
        if not xx:  # If the bit is False, then
            result += weight  # add the penalty weight to the result.
        weight += 1  # Increment the penalty weight by 1.  # noqa: SIM113

    return result  # Return the result.


class LinearHarmonic(BitStringProblem):
    """The objective function of the linear harmonic benchmark problem."""

    def __init__(self, n: int) -> None:
        """
        Initialize the Linear Harmonic problem instance.

        :param n: length of the bit string

        >>> LinearHarmonic(5).evaluate(np.array([
        ...     False, True, False, True]))
        4
        """
        super().__init__(n)
        self.evaluate = linear_harmonic  # type: ignore

    def upper_bound(self) -> int:
        """
        Return the upper bound of the linear harmonic function.

        :returns: `n * (n + 1) // 2`

        >>> LinearHarmonic(5).upper_bound()
        15
        """
        return (self.n * (self.n + 1)) // 2

    def __str__(self) -> str:
        """
        Get the name of the Linear Harmonic problem instance.

        :return: 'linearharmonic_' + n value

        >>> print(LinearHarmonic(8))
        linharm_8
        """
        return f"linharm_{self.n}"

    @classmethod
    def default_instances(
            cls: type, scale_min: int = 2, scale_max: int = 333) \
            -> Iterator[Callable[[], "LinearHarmonic"]]:
        """
        Get the 78 default instances of the :class:`LinearHarmonic` problem.

        :param scale_min: the minimum permitted scale, by default `2`
        :param scale_max: the maximum permitted scale, by default `333`
        :returns: a sequence of default :class:`LinearHarmonic` instances

        >>> len(list(LinearHarmonic.default_instances()))
        78

        >>> [x() for x in LinearHarmonic.default_instances()]
        [linharm_2, linharm_3, linharm_4, linharm_5, linharm_6, linharm_7, \
linharm_8, linharm_9, linharm_10, linharm_11, linharm_12, linharm_13, \
linharm_14, linharm_15, linharm_16, linharm_17, linharm_18, linharm_19, \
linharm_20, linharm_21, linharm_22, linharm_23, linharm_24, linharm_25, \
linharm_26, linharm_27, linharm_28, linharm_29, linharm_30, linharm_31, \
linharm_32, linharm_33, linharm_36, linharm_40, linharm_41, linharm_42, \
linharm_44, linharm_48, linharm_49, linharm_50, linharm_55, linharm_59, \
linharm_60, linharm_64, linharm_66, linharm_70, linharm_77, linharm_79, \
linharm_80, linharm_81, linharm_85, linharm_88, linharm_90, linharm_96, \
linharm_99, linharm_100, linharm_107, linharm_111, linharm_121, linharm_125, \
linharm_128, linharm_144, linharm_149, linharm_169, linharm_170, \
linharm_192, linharm_196, linharm_199, linharm_200, linharm_222, \
linharm_225, linharm_243, linharm_256, linharm_269, linharm_289, linharm_300, \
linharm_324, linharm_333]
        """
        return cast("Iterator[Callable[[], LinearHarmonic]]",
                    super().default_instances(  # type: ignore
                        scale_min, scale_max))



"""
An objective function counting the number of ones in a bit string.

This function tries to maximize the number of `True` bits in a bit string.
Therefore, it returns the lowest possible value 0 if all bits are `True`.
This is the global optimum.
It returns the highest possible value `n` if all bits are `False`. This is
the worst possible value.

1. Heinz Mühlenbein. How Genetic Algorithms Really Work: Mutation and
   Hillclimbing. In Reinhard Männer and Bernard Manderick, editors,
   *Proceedings of Parallel Problem Solving from Nature 2 (PPSN-II),*
   September 28-30, 1992, Brussels, Belgium, pages 15-26. Elsevier.
   https://www.researchgate.net/publication/220702092
2. Stefan Droste, Thomas Jansen, and Ingo Wegener. Upper and Lower Bounds for
   Randomized Search Heuristics in Black-Box Optimization. *Theory of
   Computing Systems.* 39(4):525-544. July 2006.
   doi: https://doi.org/10.1007/s00224-004-1177-z
3. Thomas Weise, Zhize Wu, Xinlu Li, and Yan Chen. Frequency Fitness
   Assignment: Making Optimization Algorithms Invariant under Bijective
   Transformations of the Objective Function Value. *IEEE Transactions on
   Evolutionary Computation* 25(2):307-319. April 2021. Preprint available at
   arXiv:2001.01416v5 [cs.NE] 15 Oct 2020.
   https://dx.doi.org/10.1109/TEVC.2020.3032090
4. v
"""

from typing import Callable, Iterator, cast

import numba  # type: ignore
import numpy as np

from moptipy.examples.bitstrings.bitstring_problem import BitStringProblem


@numba.njit(nogil=True, cache=True)
def onemax(x: np.ndarray) -> int:
    """
    Get the length of a string minus the number of ones in it.

    :param x: the np array
    :return: the length of the string minus the number of ones, i.e., the
        number of zeros

    >>> onemax(np.array([True, True, False, False, False]))
    3
    >>> onemax(np.array([True, False, True, False, False]))
    3
    >>> onemax(np.array([False, True,  True, False, False]))
    3
    >>> onemax(np.array([True, True, True, True, True]))
    0
    >>> onemax(np.array([False, True, True, True, True]))
    1
    >>> onemax(np.array([False, False, False, False, False]))
    5

    # n = 1 and 0 true bits
    >>> onemax(np.array([0]))
    1

    # n = 1 and 1 true bit
    >>> onemax(np.array([1]))
    0

    # n = 2 and 0 true bits
    >>> onemax(np.array([0, 0]))
    2

    # n = 2 and 1 true bit
    >>> onemax(np.array([0, 1]))
    1

    # n = 2 and 1 true bit
    >>> onemax(np.array([1, 0]))
    1

    # n = 2 and 1 true bit
    >>> onemax(np.array([0, 1]))
    1

    # n = 2 and 2 true bits
    >>> onemax(np.array([1, 1]))
    0

    # n = 3 and 0 true bits
    >>> onemax(np.array([0, 0, 0]))
    3

    # n = 3 and 1 true bit
    >>> onemax(np.array([1, 0, 0]))
    2

    # n = 3 and 1 true bit
    >>> onemax(np.array([0, 1, 0]))
    2

    # n = 3 and 1 true bit
    >>> onemax(np.array([0, 0, 1]))
    2

    # n = 3 and 2 true bits
    >>> onemax(np.array([1, 1, 0]))
    1

    # n = 3 and 2 true bits
    >>> onemax(np.array([0, 1, 1]))
    1

    # n = 3 and 2 true bits
    >>> onemax(np.array([1, 1, 0]))
    1

    # n = 3 and 3 true bits
    >>> onemax(np.array([1, 1, 1]))
    0

    # n = 4 and 0 true bits
    >>> onemax(np.array([0, 0, 0, 0]))
    4

    # n = 4 and 1 true bit
    >>> onemax(np.array([1, 0, 0, 0]))
    3

    # n = 4 and 1 true bit
    >>> onemax(np.array([0, 0, 0, 1]))
    3

    # n = 4 and 1 true bit
    >>> onemax(np.array([0, 0, 1, 0]))
    3

    # n = 4 and 2 true bits
    >>> onemax(np.array([0, 0, 1, 1]))
    2

    # n = 4 and 2 true bits
    >>> onemax(np.array([0, 1, 0, 1]))
    2

    # n = 4 and 2 true bits
    >>> onemax(np.array([0, 1, 0, 1]))
    2

    # n = 4 and 3 true bits
    >>> onemax(np.array([0, 1, 1, 1]))
    1

    # n = 4 and 3 true bits
    >>> onemax(np.array([1, 1, 1, 0]))
    1

    # n = 4 and 3 true bits
    >>> onemax(np.array([1, 0, 1, 1]))
    1

    # n = 4 and 4 true bits
    >>> onemax(np.array([1, 1, 1, 1]))
    0

    # n = 5 and 0 true bits
    >>> onemax(np.array([0, 0, 0, 0, 0]))
    5

    # n = 5 and 1 true bit
    >>> onemax(np.array([0, 0, 0, 0, 1]))
    4

    # n = 5 and 1 true bit
    >>> onemax(np.array([0, 1, 0, 0, 0]))
    4

    # n = 5 and 1 true bit
    >>> onemax(np.array([1, 0, 0, 0, 0]))
    4

    # n = 5 and 2 true bits
    >>> onemax(np.array([0, 1, 0, 0, 1]))
    3

    # n = 5 and 2 true bits
    >>> onemax(np.array([0, 1, 1, 0, 0]))
    3

    # n = 5 and 2 true bits
    >>> onemax(np.array([0, 0, 0, 1, 1]))
    3

    # n = 5 and 3 true bits
    >>> onemax(np.array([1, 0, 1, 1, 0]))
    2

    # n = 5 and 3 true bits
    >>> onemax(np.array([1, 1, 0, 1, 0]))
    2

    # n = 5 and 3 true bits
    >>> onemax(np.array([0, 1, 1, 1, 0]))
    2

    # n = 5 and 4 true bits
    >>> onemax(np.array([1, 0, 1, 1, 1]))
    1

    # n = 5 and 4 true bits
    >>> onemax(np.array([1, 1, 0, 1, 1]))
    1

    # n = 5 and 4 true bits
    >>> onemax(np.array([1, 0, 1, 1, 1]))
    1

    # n = 5 and 5 true bits
    >>> onemax(np.array([1, 1, 1, 1, 1]))
    0

    # n = 6 and 0 true bits
    >>> onemax(np.array([0, 0, 0, 0, 0, 0]))
    6

    # n = 6 and 1 true bit
    >>> onemax(np.array([0, 0, 0, 0, 1, 0]))
    5

    # n = 6 and 1 true bit
    >>> onemax(np.array([1, 0, 0, 0, 0, 0]))
    5

    # n = 6 and 1 true bit
    >>> onemax(np.array([0, 0, 0, 0, 1, 0]))
    5

    # n = 6 and 2 true bits
    >>> onemax(np.array([0, 1, 0, 1, 0, 0]))
    4

    # n = 6 and 2 true bits
    >>> onemax(np.array([0, 1, 0, 0, 1, 0]))
    4

    # n = 6 and 2 true bits
    >>> onemax(np.array([0, 0, 0, 0, 1, 1]))
    4

    # n = 6 and 3 true bits
    >>> onemax(np.array([0, 0, 1, 0, 1, 1]))
    3

    # n = 6 and 3 true bits
    >>> onemax(np.array([1, 0, 0, 1, 0, 1]))
    3

    # n = 6 and 3 true bits
    >>> onemax(np.array([0, 1, 1, 0, 0, 1]))
    3

    # n = 6 and 4 true bits
    >>> onemax(np.array([1, 1, 1, 0, 1, 0]))
    2

    # n = 6 and 4 true bits
    >>> onemax(np.array([1, 1, 0, 0, 1, 1]))
    2

    # n = 6 and 4 true bits
    >>> onemax(np.array([1, 1, 1, 0, 1, 0]))
    2

    # n = 6 and 5 true bits
    >>> onemax(np.array([1, 1, 0, 1, 1, 1]))
    1

    # n = 6 and 5 true bits
    >>> onemax(np.array([1, 1, 1, 1, 0, 1]))
    1

    # n = 6 and 5 true bits
    >>> onemax(np.array([1, 1, 1, 1, 0, 1]))
    1

    # n = 6 and 6 true bits
    >>> onemax(np.array([1, 1, 1, 1, 1, 1]))
    0

    # n = 7 and 0 true bits
    >>> onemax(np.array([0, 0, 0, 0, 0, 0, 0]))
    7

    # n = 7 and 1 true bit
    >>> onemax(np.array([0, 1, 0, 0, 0, 0, 0]))
    6

    # n = 7 and 1 true bit
    >>> onemax(np.array([0, 0, 0, 0, 1, 0, 0]))
    6

    # n = 7 and 1 true bit
    >>> onemax(np.array([0, 1, 0, 0, 0, 0, 0]))
    6

    # n = 7 and 2 true bits
    >>> onemax(np.array([1, 0, 0, 0, 1, 0, 0]))
    5

    # n = 7 and 2 true bits
    >>> onemax(np.array([0, 1, 0, 1, 0, 0, 0]))
    5

    # n = 7 and 2 true bits
    >>> onemax(np.array([1, 0, 0, 0, 0, 0, 1]))
    5

    # n = 7 and 3 true bits
    >>> onemax(np.array([1, 0, 1, 1, 0, 0, 0]))
    4

    # n = 7 and 3 true bits
    >>> onemax(np.array([0, 0, 1, 1, 0, 0, 1]))
    4

    # n = 7 and 3 true bits
    >>> onemax(np.array([0, 0, 1, 1, 0, 0, 1]))
    4

    # n = 7 and 4 true bits
    >>> onemax(np.array([0, 1, 0, 1, 1, 1, 0]))
    3

    # n = 7 and 4 true bits
    >>> onemax(np.array([1, 1, 1, 0, 1, 0, 0]))
    3

    # n = 7 and 4 true bits
    >>> onemax(np.array([0, 1, 1, 1, 0, 1, 0]))
    3

    # n = 7 and 5 true bits
    >>> onemax(np.array([1, 1, 1, 1, 1, 0, 0]))
    2

    # n = 7 and 5 true bits
    >>> onemax(np.array([1, 1, 1, 1, 0, 1, 0]))
    2

    # n = 7 and 5 true bits
    >>> onemax(np.array([0, 1, 1, 1, 1, 1, 0]))
    2

    # n = 7 and 6 true bits
    >>> onemax(np.array([1, 1, 1, 1, 0, 1, 1]))
    1

    # n = 7 and 6 true bits
    >>> onemax(np.array([1, 1, 1, 0, 1, 1, 1]))
    1

    # n = 7 and 6 true bits
    >>> onemax(np.array([0, 1, 1, 1, 1, 1, 1]))
    1

    # n = 7 and 7 true bits
    >>> onemax(np.array([1, 1, 1, 1, 1, 1, 1]))
    0

    # n = 8 and 0 true bits
    >>> onemax(np.array([0, 0, 0, 0, 0, 0, 0, 0]))
    8

    # n = 8 and 1 true bit
    >>> onemax(np.array([0, 0, 0, 0, 0, 1, 0, 0]))
    7

    # n = 8 and 1 true bit
    >>> onemax(np.array([1, 0, 0, 0, 0, 0, 0, 0]))
    7

    # n = 8 and 1 true bit
    >>> onemax(np.array([0, 0, 0, 0, 0, 0, 0, 1]))
    7

    # n = 8 and 2 true bits
    >>> onemax(np.array([0, 0, 0, 0, 0, 0, 1, 1]))
    6

    # n = 8 and 2 true bits
    >>> onemax(np.array([0, 0, 1, 0, 0, 1, 0, 0]))
    6

    # n = 8 and 2 true bits
    >>> onemax(np.array([0, 0, 0, 0, 0, 1, 1, 0]))
    6

    # n = 8 and 3 true bits
    >>> onemax(np.array([0, 1, 0, 0, 1, 1, 0, 0]))
    5

    # n = 8 and 3 true bits
    >>> onemax(np.array([0, 1, 0, 1, 1, 0, 0, 0]))
    5

    # n = 8 and 3 true bits
    >>> onemax(np.array([0, 0, 0, 0, 1, 1, 1, 0]))
    5

    # n = 8 and 4 true bits
    >>> onemax(np.array([1, 0, 1, 1, 0, 0, 1, 0]))
    4

    # n = 8 and 4 true bits
    >>> onemax(np.array([1, 1, 0, 0, 1, 1, 0, 0]))
    4

    # n = 8 and 4 true bits
    >>> onemax(np.array([1, 0, 0, 0, 1, 1, 0, 1]))
    4

    # n = 8 and 5 true bits
    >>> onemax(np.array([0, 1, 1, 0, 0, 1, 1, 1]))
    3

    # n = 8 and 5 true bits
    >>> onemax(np.array([1, 1, 0, 0, 1, 0, 1, 1]))
    3

    # n = 8 and 5 true bits
    >>> onemax(np.array([1, 1, 0, 0, 1, 1, 0, 1]))
    3

    # n = 8 and 6 true bits
    >>> onemax(np.array([1, 1, 0, 1, 1, 1, 1, 0]))
    2

    # n = 8 and 6 true bits
    >>> onemax(np.array([1, 1, 0, 1, 1, 0, 1, 1]))
    2

    # n = 8 and 6 true bits
    >>> onemax(np.array([0, 0, 1, 1, 1, 1, 1, 1]))
    2

    # n = 8 and 7 true bits
    >>> onemax(np.array([1, 1, 1, 1, 1, 1, 0, 1]))
    1

    # n = 8 and 7 true bits
    >>> onemax(np.array([1, 1, 1, 1, 1, 1, 1, 0]))
    1

    # n = 8 and 7 true bits
    >>> onemax(np.array([1, 1, 1, 1, 1, 0, 1, 1]))
    1

    # n = 8 and 8 true bits
    >>> onemax(np.array([1, 1, 1, 1, 1, 1, 1, 1]))
    0

    # n = 9 and 0 true bits
    >>> onemax(np.array([0, 0, 0, 0, 0, 0, 0, 0, 0]))
    9

    # n = 9 and 1 true bit
    >>> onemax(np.array([0, 0, 0, 0, 0, 0, 0, 1, 0]))
    8

    # n = 9 and 1 true bit
    >>> onemax(np.array([0, 0, 0, 0, 0, 0, 0, 0, 1]))
    8

    # n = 9 and 1 true bit
    >>> onemax(np.array([0, 0, 0, 0, 1, 0, 0, 0, 0]))
    8

    # n = 9 and 2 true bits
    >>> onemax(np.array([1, 0, 0, 0, 0, 0, 0, 1, 0]))
    7

    # n = 9 and 2 true bits
    >>> onemax(np.array([0, 1, 0, 0, 0, 0, 1, 0, 0]))
    7

    # n = 9 and 2 true bits
    >>> onemax(np.array([0, 0, 1, 0, 1, 0, 0, 0, 0]))
    7

    # n = 9 and 3 true bits
    >>> onemax(np.array([0, 0, 1, 0, 0, 1, 0, 0, 1]))
    6

    # n = 9 and 3 true bits
    >>> onemax(np.array([1, 0, 0, 0, 0, 0, 0, 1, 1]))
    6

    # n = 9 and 3 true bits
    >>> onemax(np.array([0, 0, 1, 0, 1, 0, 0, 1, 0]))
    6

    # n = 9 and 4 true bits
    >>> onemax(np.array([0, 0, 0, 0, 1, 1, 0, 1, 1]))
    5

    # n = 9 and 4 true bits
    >>> onemax(np.array([0, 0, 0, 1, 0, 0, 1, 1, 1]))
    5

    # n = 9 and 4 true bits
    >>> onemax(np.array([0, 0, 1, 1, 1, 0, 0, 0, 1]))
    5

    # n = 9 and 5 true bits
    >>> onemax(np.array([0, 0, 1, 1, 0, 1, 0, 1, 1]))
    4

    # n = 9 and 5 true bits
    >>> onemax(np.array([0, 0, 1, 1, 0, 1, 0, 1, 1]))
    4

    # n = 9 and 5 true bits
    >>> onemax(np.array([1, 1, 0, 0, 1, 0, 0, 1, 1]))
    4

    # n = 9 and 6 true bits
    >>> onemax(np.array([0, 0, 1, 1, 1, 1, 1, 0, 1]))
    3

    # n = 9 and 6 true bits
    >>> onemax(np.array([1, 1, 0, 0, 1, 0, 1, 1, 1]))
    3

    # n = 9 and 6 true bits
    >>> onemax(np.array([0, 1, 1, 1, 0, 1, 1, 0, 1]))
    3

    # n = 9 and 7 true bits
    >>> onemax(np.array([0, 1, 1, 0, 1, 1, 1, 1, 1]))
    2

    # n = 9 and 7 true bits
    >>> onemax(np.array([1, 1, 0, 1, 1, 1, 1, 0, 1]))
    2

    # n = 9 and 7 true bits
    >>> onemax(np.array([1, 1, 1, 0, 1, 1, 0, 1, 1]))
    2

    # n = 9 and 8 true bits
    >>> onemax(np.array([1, 1, 1, 0, 1, 1, 1, 1, 1]))
    1

    # n = 9 and 8 true bits
    >>> onemax(np.array([1, 1, 1, 1, 1, 0, 1, 1, 1]))
    1

    # n = 9 and 8 true bits
    >>> onemax(np.array([1, 1, 1, 0, 1, 1, 1, 1, 1]))
    1

    # n = 9 and 9 true bits
    >>> onemax(np.array([1, 1, 1, 1, 1, 1, 1, 1, 1]))
    0

    # n = 10 and 0 true bits
    >>> onemax(np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0]))
    10

    # n = 10 and 1 true bit
    >>> onemax(np.array([0, 0, 0, 1, 0, 0, 0, 0, 0, 0]))
    9

    # n = 10 and 1 true bit
    >>> onemax(np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 1]))
    9

    # n = 10 and 1 true bit
    >>> onemax(np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 1]))
    9

    # n = 10 and 2 true bits
    >>> onemax(np.array([0, 1, 0, 0, 0, 0, 0, 1, 0, 0]))
    8

    # n = 10 and 2 true bits
    >>> onemax(np.array([1, 1, 0, 0, 0, 0, 0, 0, 0, 0]))
    8

    # n = 10 and 2 true bits
    >>> onemax(np.array([1, 0, 0, 1, 0, 0, 0, 0, 0, 0]))
    8

    # n = 10 and 3 true bits
    >>> onemax(np.array([0, 0, 0, 1, 0, 0, 1, 1, 0, 0]))
    7

    # n = 10 and 3 true bits
    >>> onemax(np.array([1, 0, 0, 0, 0, 0, 1, 0, 1, 0]))
    7

    # n = 10 and 3 true bits
    >>> onemax(np.array([0, 1, 0, 0, 0, 0, 1, 1, 0, 0]))
    7

    # n = 10 and 4 true bits
    >>> onemax(np.array([0, 0, 1, 1, 1, 0, 0, 1, 0, 0]))
    6

    # n = 10 and 4 true bits
    >>> onemax(np.array([0, 1, 1, 0, 0, 1, 0, 0, 0, 1]))
    6

    # n = 10 and 4 true bits
    >>> onemax(np.array([0, 0, 1, 1, 0, 1, 0, 0, 1, 0]))
    6

    # n = 10 and 5 true bits
    >>> onemax(np.array([1, 1, 1, 0, 0, 1, 0, 1, 0, 0]))
    5

    # n = 10 and 5 true bits
    >>> onemax(np.array([0, 1, 1, 0, 1, 1, 0, 0, 0, 1]))
    5

    # n = 10 and 5 true bits
    >>> onemax(np.array([0, 1, 0, 1, 0, 1, 0, 1, 0, 1]))
    5

    # n = 10 and 6 true bits
    >>> onemax(np.array([1, 1, 1, 0, 1, 0, 1, 1, 0, 0]))
    4

    # n = 10 and 6 true bits
    >>> onemax(np.array([1, 1, 1, 1, 1, 0, 1, 0, 0, 0]))
    4

    # n = 10 and 6 true bits
    >>> onemax(np.array([0, 0, 1, 1, 1, 1, 0, 1, 0, 1]))
    4

    # n = 10 and 7 true bits
    >>> onemax(np.array([1, 1, 1, 1, 0, 0, 1, 1, 0, 1]))
    3

    # n = 10 and 7 true bits
    >>> onemax(np.array([1, 1, 0, 1, 1, 1, 0, 1, 1, 0]))
    3

    # n = 10 and 7 true bits
    >>> onemax(np.array([0, 0, 1, 1, 1, 1, 0, 1, 1, 1]))
    3

    # n = 10 and 8 true bits
    >>> onemax(np.array([0, 1, 1, 1, 0, 1, 1, 1, 1, 1]))
    2

    # n = 10 and 8 true bits
    >>> onemax(np.array([1, 1, 1, 1, 1, 0, 1, 1, 1, 0]))
    2

    # n = 10 and 8 true bits
    >>> onemax(np.array([1, 1, 1, 1, 0, 1, 0, 1, 1, 1]))
    2

    # n = 10 and 9 true bits
    >>> onemax(np.array([1, 1, 1, 1, 1, 1, 1, 1, 0, 1]))
    1

    # n = 10 and 9 true bits
    >>> onemax(np.array([0, 1, 1, 1, 1, 1, 1, 1, 1, 1]))
    1

    # n = 10 and 9 true bits
    >>> onemax(np.array([1, 1, 1, 1, 1, 1, 0, 1, 1, 1]))
    1

    # n = 10 and 10 true bits
    >>> onemax(np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1]))
    0
    """
    return len(x) - int(x.sum())


class OneMax(BitStringProblem):
    """Maximize the number of ones in a bit string."""

    def __init__(self, n: int) -> None:
        """
        Initialize the onemax objective function.

        :param n: the dimension of the problem

        >>> print(OneMax(2).n)
        2
        >>> print(OneMax(4).evaluate(np.array([True, True, False, True])))
        1
        """
        super().__init__(n)
        self.evaluate = onemax  # type: ignore

    def __str__(self) -> str:
        """
        Get the name of the onemax objective function.

        :return: `onemax_` + length of string

        >>> OneMax(13)
        onemax_13
        """
        return f"onemax_{self.n}"

    @classmethod
    def default_instances(
            cls: type, scale_min: int = 2, scale_max: int = 8192) \
            -> Iterator[Callable[[], "OneMax"]]:
        """
        Get the 202 default instances of the :class:`OneMax` problem.

        :param scale_min: the minimum permitted scale, by default `2`
        :param scale_max: the maximum permitted scale, by default `8192`
        :returns: a sequence of default :class:`OneMax` instances

        >>> len(list(OneMax.default_instances()))
        202

        >>> [x() for x in OneMax.default_instances()]
        [onemax_2, onemax_3, onemax_4, onemax_5, onemax_6, onemax_7, \
onemax_8, onemax_9, onemax_10, onemax_11, onemax_12, onemax_13, onemax_14, \
onemax_15, onemax_16, onemax_17, onemax_18, onemax_19, onemax_20, onemax_21, \
onemax_22, onemax_23, onemax_24, onemax_25, onemax_26, onemax_27, onemax_28, \
onemax_29, onemax_30, onemax_31, onemax_32, onemax_33, onemax_36, onemax_40, \
onemax_41, onemax_42, onemax_44, onemax_48, onemax_49, onemax_50, onemax_55, \
onemax_59, onemax_60, onemax_64, onemax_66, onemax_70, onemax_77, onemax_79, \
onemax_80, onemax_81, onemax_85, onemax_88, onemax_90, onemax_96, onemax_99, \
onemax_100, onemax_107, onemax_111, onemax_121, onemax_125, onemax_128, \
onemax_144, onemax_149, onemax_169, onemax_170, onemax_192, onemax_196, \
onemax_199, onemax_200, onemax_222, onemax_225, onemax_243, onemax_256, \
onemax_269, onemax_289, onemax_300, onemax_324, onemax_333, onemax_341, \
onemax_343, onemax_359, onemax_361, onemax_384, onemax_400, onemax_441, \
onemax_444, onemax_479, onemax_484, onemax_500, onemax_512, onemax_529, \
onemax_555, onemax_576, onemax_600, onemax_625, onemax_641, onemax_666, \
onemax_676, onemax_682, onemax_700, onemax_729, onemax_768, onemax_777, \
onemax_784, onemax_800, onemax_841, onemax_857, onemax_888, onemax_900, \
onemax_961, onemax_999, onemax_1000, onemax_1024, onemax_1089, onemax_1111, \
onemax_1151, onemax_1156, onemax_1225, onemax_1296, onemax_1365, \
onemax_1369, onemax_1444, onemax_1521, onemax_1536, onemax_1543, \
onemax_1600, onemax_1681, onemax_1764, onemax_1849, onemax_1936, \
onemax_2000, onemax_2025, onemax_2048, onemax_2063, onemax_2116, \
onemax_2187, onemax_2209, onemax_2222, onemax_2304, onemax_2401, \
onemax_2500, onemax_2601, onemax_2704, onemax_2730, onemax_2753, \
onemax_2809, onemax_2916, onemax_3000, onemax_3025, onemax_3072, \
onemax_3125, onemax_3136, onemax_3249, onemax_3333, onemax_3364, \
onemax_3481, onemax_3600, onemax_3671, onemax_3721, onemax_3844, \
onemax_3969, onemax_4000, onemax_4096, onemax_4225, onemax_4356, \
onemax_4444, onemax_4489, onemax_4624, onemax_4761, onemax_4900, \
onemax_4903, onemax_5000, onemax_5041, onemax_5184, onemax_5329, \
onemax_5461, onemax_5476, onemax_5555, onemax_5625, onemax_5776, \
onemax_5929, onemax_6000, onemax_6084, onemax_6144, onemax_6241, \
onemax_6400, onemax_6547, onemax_6561, onemax_6666, onemax_6724, \
onemax_6889, onemax_7000, onemax_7056, onemax_7225, onemax_7396, \
onemax_7569, onemax_7744, onemax_7777, onemax_7921, onemax_8000, \
onemax_8100, onemax_8192]

        """
        return cast("Iterator[Callable[[], OneMax]]",
                    super().default_instances(  # type: ignore
                        scale_min, scale_max))



"""
The SAFEA-B is hybrid of the (1+1)FEA and the SA with Solution Transfer.

The algorithm combines frequency fitness assignment based local search, i.e.,
the FEA, with simulated annealing (SA). Both algorithms get assigned
alternating objective function evaluations (FEs). The FEA branch remains
unchanged, it is never disturbed and no information flows from the simulated
annealing branch over to it. However, solutions are copied from time to time
from the FEA branch to the SA branch. The solution is transferred from the FEA
branch to the SA branch if its better than the current solution in that
branch.

1. Tianyu Liang, Zhize Wu, Jörg Lässig, Daan van den Berg, Sarah Louise
   Thomson, and Thomas Weise. Addressing the Traveling Salesperson Problem
   with Frequency Fitness Assignment and Hybrid Algorithms. *Soft Computing.*
   2024. https://dx.doi.org/10.1007/s00500-024-09718-8
"""
from collections import Counter
from math import exp
from typing import Callable, Final

from numpy.random import Generator
from pycommons.types import type_error

from moptipy.algorithms.modules.temperature_schedule import TemperatureSchedule
from moptipy.algorithms.so.ffa.ffa_h import create_h, log_h
from moptipy.api.algorithm import Algorithm1
from moptipy.api.operators import Op0, Op1
from moptipy.api.process import Process
from moptipy.utils.logger import KeyValueLogSection


class SAFEAB(Algorithm1):
    """An implementation of the SAFEA-B."""

    def __init__(self, op0: Op0, op1: Op1, schedule: TemperatureSchedule,
                 log_h_tbl: bool = False) -> None:
        """
        Create the SAFEA-B.

        :param op0: the nullary search operator
        :param op1: the unary search operator
        :param schedule: the temperature schedule to use
        :param log_h_tbl: should we log the H table?
        """
        if not isinstance(schedule, TemperatureSchedule):
            raise type_error(schedule, "schedule", TemperatureSchedule)
        if not isinstance(log_h_tbl, bool):
            raise type_error(log_h_tbl, "log_h_tbl", bool)
        super().__init__(f"safeaB_{schedule}", op0, op1)
        #: True if we should log the H table, False otherwise
        self.__log_h_tbl: Final[bool] = log_h_tbl
        #: the temperature schedule
        self.schedule: Final[TemperatureSchedule] = schedule

    def solve(self, process: Process) -> None:
        """
        Apply the SAFEA-B to an optimization problem.

        :param process: the black-box process object
        """
        # Create records for old and new point in the search space.
        x_c = process.create()  # record for current solution of the SA
        x_d = process.create()  # record for current solution of the FEA
        x_n = process.create()  # record for new solution

        # Obtain the random number generator.
        random: Final[Generator] = process.get_random()

        # Put function references in variables to save time.
        evaluate: Final[Callable] = process.evaluate  # the objective
        should_terminate: Final[Callable] = process.should_terminate
        temperature: Final[Callable[[int], float]] = self.schedule.temperature
        r01: Final[Callable[[], float]] = random.random  # random from [0, 1)
        xcopy: Final[Callable] = process.copy  # copy(dest, source)
        op0: Final[Callable] = self.op0.op0  # the nullary operator
        op1: Final[Callable] = self.op1.op1  # the unary operator

        h, ofs = create_h(process)  # Allocate the h-table

        # Start at a random point in the search space and evaluate it.
        op0(random, x_c)  # Create 1 solution randomly and
        y_c: int | float = evaluate(x_c) + ofs  # evaluate it.
        xcopy(x_d, x_c)
        y_d: int | float = y_c
        use_ffa: bool = True
        tau: int = 0  # The iteration index, needs to be 0 at first cmp.

        while not should_terminate():  # Until we need to quit...
            use_ffa = not use_ffa  # toggle use of FFA
            op1(random, x_n, x_d if use_ffa else x_c)
            y_n: int | float = evaluate(x_n) + ofs

            if use_ffa:  # the FEA branch
                h[y_n] += 1  # type: ignore  # Increase the frequency
                h[y_d] += 1  # type: ignore  # of new_f and cur_f.
                if h[y_n] <= h[y_d]:  # type: ignore
                    y_d = y_n  # Store its objective value.
                    x_d, x_n = x_n, x_d  # Swap best and new.
                    if y_n <= y_c:  # check transfer if solution was
                        y_c = y_n   # accepted
                        xcopy(x_c, x_d)  # copy the solution over
                    continue  # skip rest of loop body
            if (y_n <= y_c) or (  # Accept if <= or if SA criterion
                    (not use_ffa) and (  # Only for SA, not FEA
                        r01() < exp((y_c - y_n) / temperature(tau)))):
                y_c = y_n
                x_c, x_n = x_n, x_c
            tau += 1  # Step the iteration index.

        if not self.__log_h_tbl:
            return  # we are done here

        # After we are done, we want to print the H-table.
        if h[y_c] == 0:  # type: ignore  # Fix the H-table for the case
            h = Counter()   # that only one FE was performed: In this case,
            h[y_c] = 1  # make Counter with only a single 1 value inside.

        log_h(process, h, ofs)  # log the H-table

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of the SAFEA-B algorithm.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        with logger.scope("ts") as ts:
            self.schedule.log_parameters_to(ts)



"""
The one-dimensional Ising problem.

The one-dimensional Ising problem describes a ring. For each bit that differs
from its (right-side) neighboring bit, a penalty of 1 is incurred. The optimum
is a bit string of either all ones or all zeros. The optimal objective value
is 0, the worst-possible one is `n`.

1. Simon Fischer and Ingo Wegener. The one-dimensional Ising model: Mutation
   versus recombination. *Theoretical Computer Science.* 344(2-3):208-225.
   November 2005. doi: https://doi.org/10.1016/j.tcs.2005.04.002
2. Carola Doerr and Furong Ye and Naama Horesh and Hao Wang and Ofer M. Shir
   and Thomas Bäck. Benchmarking Discrete Optimization Heuristics with
   IOHprofiler. Applied Soft Computing 88:106027, March 2020,
   doi: https://doi.org/10.1016/j.asoc.2019.106027.
3. Clarissa Van Hoyweghen, David Edward Goldberg, and Bart Naudts. From Twomax
   To The Ising Model: Easy And Hard Symmetrical Problems. *In Proceedings of
   the Genetic and Evolutionary Computation Conference (GECCO'02),* July 9-13,
   2002, New York, NY, USA, pages 626-633. Morgan Kaufmann.
   http://gpbib.cs.ucl.ac.uk/gecco2002/GA013.pdf
4. Thomas Weise, Zhize Wu, Xinlu Li, Yan Chen, and Jörg Lässig. Frequency
   Fitness Assignment: Optimization without Bias for Good Solutions can be
   Efficient. *IEEE Transactions on Evolutionary Computation (TEVC)*.
   27(4):980-992. August 2023.
   doi: https://doi.org/10.1109/TEVC.2022.3191698
"""

from typing import Callable, Iterator, cast

import numba  # type: ignore
import numpy as np

from moptipy.examples.bitstrings.bitstring_problem import BitStringProblem


@numba.njit(nogil=True, cache=True)
def ising1d(x: np.ndarray) -> int:
    """
    Compute the objective value of the 1-dimensional Ising problem.

    :param x: the np array
    :return: the ising1d function value

    >>> ising1d(np.array([True, True, True, True, True]))
    0
    >>> ising1d(np.array([False, False, False, False, False]))
    0
    >>> ising1d(np.array([False, False, False, True, False]))
    2
    >>> ising1d(np.array([True, False, False, False, False]))
    2
    >>> ising1d(np.array([False, False, False, False, True]))
    2
    >>> ising1d(np.array([True, False, False, False, True]))
    2
    >>> ising1d(np.array([True, False, True, False, False]))
    4
    >>> ising1d(np.array([True, False, True, False, True, False]))
    6

    # n = 1 and 0 true bits
    >>> ising1d(np.array([0]))
    0

    # n = 1 and 1 true bit
    >>> ising1d(np.array([1]))
    0

    # n = 2 and 0 true bits
    >>> ising1d(np.array([0, 0]))
    0

    # n = 2 and 1 true bit
    >>> ising1d(np.array([1, 0]))
    2

    # n = 2 and 1 true bit
    >>> ising1d(np.array([0, 1]))
    2

    # n = 2 and 1 true bit
    >>> ising1d(np.array([0, 1]))
    2

    # n = 2 and 2 true bits
    >>> ising1d(np.array([1, 1]))
    0

    # n = 3 and 0 true bits
    >>> ising1d(np.array([0, 0, 0]))
    0

    # n = 3 and 1 true bit
    >>> ising1d(np.array([1, 0, 0]))
    2

    # n = 3 and 1 true bit
    >>> ising1d(np.array([0, 0, 1]))
    2

    # n = 3 and 1 true bit
    >>> ising1d(np.array([1, 0, 0]))
    2

    # n = 3 and 2 true bits
    >>> ising1d(np.array([1, 0, 1]))
    2

    # n = 3 and 2 true bits
    >>> ising1d(np.array([1, 0, 1]))
    2

    # n = 3 and 2 true bits
    >>> ising1d(np.array([1, 1, 0]))
    2

    # n = 3 and 3 true bits
    >>> ising1d(np.array([1, 1, 1]))
    0

    # n = 4 and 0 true bits
    >>> ising1d(np.array([0, 0, 0, 0]))
    0

    # n = 4 and 1 true bit
    >>> ising1d(np.array([1, 0, 0, 0]))
    2

    # n = 4 and 1 true bit
    >>> ising1d(np.array([0, 0, 1, 0]))
    2

    # n = 4 and 1 true bit
    >>> ising1d(np.array([1, 0, 0, 0]))
    2

    # n = 4 and 2 true bits
    >>> ising1d(np.array([1, 0, 0, 1]))
    2

    # n = 4 and 2 true bits
    >>> ising1d(np.array([1, 1, 0, 0]))
    2

    # n = 4 and 2 true bits
    >>> ising1d(np.array([0, 1, 1, 0]))
    2

    # n = 4 and 3 true bits
    >>> ising1d(np.array([0, 1, 1, 1]))
    2

    # n = 4 and 3 true bits
    >>> ising1d(np.array([1, 1, 0, 1]))
    2

    # n = 4 and 3 true bits
    >>> ising1d(np.array([0, 1, 1, 1]))
    2

    # n = 4 and 4 true bits
    >>> ising1d(np.array([1, 1, 1, 1]))
    0

    # n = 5 and 0 true bits
    >>> ising1d(np.array([0, 0, 0, 0, 0]))
    0

    # n = 5 and 1 true bit
    >>> ising1d(np.array([0, 0, 0, 0, 1]))
    2

    # n = 5 and 1 true bit
    >>> ising1d(np.array([0, 0, 0, 0, 1]))
    2

    # n = 5 and 1 true bit
    >>> ising1d(np.array([0, 1, 0, 0, 0]))
    2

    # n = 5 and 2 true bits
    >>> ising1d(np.array([0, 1, 0, 0, 1]))
    4

    # n = 5 and 2 true bits
    >>> ising1d(np.array([0, 1, 0, 0, 1]))
    4

    # n = 5 and 2 true bits
    >>> ising1d(np.array([0, 1, 1, 0, 0]))
    2

    # n = 5 and 3 true bits
    >>> ising1d(np.array([1, 1, 0, 1, 0]))
    4

    # n = 5 and 3 true bits
    >>> ising1d(np.array([1, 1, 1, 0, 0]))
    2

    # n = 5 and 3 true bits
    >>> ising1d(np.array([1, 0, 1, 0, 1]))
    4

    # n = 5 and 4 true bits
    >>> ising1d(np.array([1, 1, 1, 1, 0]))
    2

    # n = 5 and 4 true bits
    >>> ising1d(np.array([1, 1, 0, 1, 1]))
    2

    # n = 5 and 4 true bits
    >>> ising1d(np.array([1, 1, 0, 1, 1]))
    2

    # n = 5 and 5 true bits
    >>> ising1d(np.array([1, 1, 1, 1, 1]))
    0

    # n = 6 and 0 true bits
    >>> ising1d(np.array([0, 0, 0, 0, 0, 0]))
    0

    # n = 6 and 1 true bit
    >>> ising1d(np.array([0, 0, 0, 0, 0, 1]))
    2

    # n = 6 and 1 true bit
    >>> ising1d(np.array([0, 0, 0, 0, 1, 0]))
    2

    # n = 6 and 1 true bit
    >>> ising1d(np.array([0, 1, 0, 0, 0, 0]))
    2

    # n = 6 and 2 true bits
    >>> ising1d(np.array([1, 1, 0, 0, 0, 0]))
    2

    # n = 6 and 2 true bits
    >>> ising1d(np.array([1, 1, 0, 0, 0, 0]))
    2

    # n = 6 and 2 true bits
    >>> ising1d(np.array([0, 0, 0, 1, 1, 0]))
    2

    # n = 6 and 3 true bits
    >>> ising1d(np.array([1, 0, 0, 1, 0, 1]))
    4

    # n = 6 and 3 true bits
    >>> ising1d(np.array([1, 0, 0, 0, 1, 1]))
    2

    # n = 6 and 3 true bits
    >>> ising1d(np.array([1, 1, 0, 0, 1, 0]))
    4

    # n = 6 and 4 true bits
    >>> ising1d(np.array([1, 0, 1, 1, 1, 0]))
    4

    # n = 6 and 4 true bits
    >>> ising1d(np.array([1, 1, 1, 1, 0, 0]))
    2

    # n = 6 and 4 true bits
    >>> ising1d(np.array([1, 1, 0, 1, 0, 1]))
    4

    # n = 6 and 5 true bits
    >>> ising1d(np.array([1, 1, 1, 1, 0, 1]))
    2

    # n = 6 and 5 true bits
    >>> ising1d(np.array([1, 0, 1, 1, 1, 1]))
    2

    # n = 6 and 5 true bits
    >>> ising1d(np.array([0, 1, 1, 1, 1, 1]))
    2

    # n = 6 and 6 true bits
    >>> ising1d(np.array([1, 1, 1, 1, 1, 1]))
    0

    # n = 7 and 0 true bits
    >>> ising1d(np.array([0, 0, 0, 0, 0, 0, 0]))
    0

    # n = 7 and 1 true bit
    >>> ising1d(np.array([0, 0, 0, 1, 0, 0, 0]))
    2

    # n = 7 and 1 true bit
    >>> ising1d(np.array([0, 0, 0, 0, 0, 1, 0]))
    2

    # n = 7 and 1 true bit
    >>> ising1d(np.array([0, 0, 0, 0, 0, 1, 0]))
    2

    # n = 7 and 2 true bits
    >>> ising1d(np.array([0, 1, 1, 0, 0, 0, 0]))
    2

    # n = 7 and 2 true bits
    >>> ising1d(np.array([1, 0, 0, 0, 0, 1, 0]))
    4

    # n = 7 and 2 true bits
    >>> ising1d(np.array([1, 0, 0, 0, 0, 0, 1]))
    2

    # n = 7 and 3 true bits
    >>> ising1d(np.array([1, 0, 0, 0, 0, 1, 1]))
    2

    # n = 7 and 3 true bits
    >>> ising1d(np.array([1, 0, 1, 0, 0, 1, 0]))
    6

    # n = 7 and 3 true bits
    >>> ising1d(np.array([1, 0, 0, 1, 0, 1, 0]))
    6

    # n = 7 and 4 true bits
    >>> ising1d(np.array([0, 1, 1, 0, 1, 1, 0]))
    4

    # n = 7 and 4 true bits
    >>> ising1d(np.array([0, 1, 0, 1, 1, 0, 1]))
    6

    # n = 7 and 4 true bits
    >>> ising1d(np.array([1, 1, 1, 0, 1, 0, 0]))
    4

    # n = 7 and 5 true bits
    >>> ising1d(np.array([1, 0, 1, 1, 1, 0, 1]))
    4

    # n = 7 and 5 true bits
    >>> ising1d(np.array([0, 1, 1, 1, 0, 1, 1]))
    4

    # n = 7 and 5 true bits
    >>> ising1d(np.array([1, 1, 1, 0, 1, 1, 0]))
    4

    # n = 7 and 6 true bits
    >>> ising1d(np.array([1, 1, 1, 0, 1, 1, 1]))
    2

    # n = 7 and 6 true bits
    >>> ising1d(np.array([1, 1, 1, 1, 1, 1, 0]))
    2

    # n = 7 and 6 true bits
    >>> ising1d(np.array([0, 1, 1, 1, 1, 1, 1]))
    2

    # n = 7 and 7 true bits
    >>> ising1d(np.array([1, 1, 1, 1, 1, 1, 1]))
    0

    # n = 8 and 0 true bits
    >>> ising1d(np.array([0, 0, 0, 0, 0, 0, 0, 0]))
    0

    # n = 8 and 1 true bit
    >>> ising1d(np.array([0, 0, 1, 0, 0, 0, 0, 0]))
    2

    # n = 8 and 1 true bit
    >>> ising1d(np.array([0, 0, 1, 0, 0, 0, 0, 0]))
    2

    # n = 8 and 1 true bit
    >>> ising1d(np.array([0, 0, 0, 0, 0, 1, 0, 0]))
    2

    # n = 8 and 2 true bits
    >>> ising1d(np.array([1, 0, 1, 0, 0, 0, 0, 0]))
    4

    # n = 8 and 2 true bits
    >>> ising1d(np.array([0, 0, 0, 0, 1, 1, 0, 0]))
    2

    # n = 8 and 2 true bits
    >>> ising1d(np.array([0, 0, 0, 1, 0, 0, 0, 1]))
    4

    # n = 8 and 3 true bits
    >>> ising1d(np.array([0, 0, 1, 1, 1, 0, 0, 0]))
    2

    # n = 8 and 3 true bits
    >>> ising1d(np.array([1, 0, 1, 0, 0, 0, 1, 0]))
    6

    # n = 8 and 3 true bits
    >>> ising1d(np.array([0, 1, 0, 1, 0, 0, 1, 0]))
    6

    # n = 8 and 4 true bits
    >>> ising1d(np.array([0, 1, 0, 1, 1, 0, 0, 1]))
    6

    # n = 8 and 4 true bits
    >>> ising1d(np.array([1, 0, 1, 0, 1, 0, 1, 0]))
    8

    # n = 8 and 4 true bits
    >>> ising1d(np.array([1, 0, 1, 0, 1, 0, 0, 1]))
    6

    # n = 8 and 5 true bits
    >>> ising1d(np.array([1, 0, 1, 0, 0, 1, 1, 1]))
    4

    # n = 8 and 5 true bits
    >>> ising1d(np.array([1, 1, 0, 1, 0, 0, 1, 1]))
    4

    # n = 8 and 5 true bits
    >>> ising1d(np.array([1, 0, 0, 1, 0, 1, 1, 1]))
    4

    # n = 8 and 6 true bits
    >>> ising1d(np.array([0, 1, 1, 1, 0, 1, 1, 1]))
    4

    # n = 8 and 6 true bits
    >>> ising1d(np.array([0, 0, 1, 1, 1, 1, 1, 1]))
    2

    # n = 8 and 6 true bits
    >>> ising1d(np.array([1, 1, 1, 1, 1, 1, 0, 0]))
    2

    # n = 8 and 7 true bits
    >>> ising1d(np.array([1, 1, 1, 0, 1, 1, 1, 1]))
    2

    # n = 8 and 7 true bits
    >>> ising1d(np.array([1, 1, 1, 1, 1, 1, 0, 1]))
    2

    # n = 8 and 7 true bits
    >>> ising1d(np.array([1, 1, 1, 1, 1, 1, 0, 1]))
    2

    # n = 8 and 8 true bits
    >>> ising1d(np.array([1, 1, 1, 1, 1, 1, 1, 1]))
    0

    # n = 9 and 0 true bits
    >>> ising1d(np.array([0, 0, 0, 0, 0, 0, 0, 0, 0]))
    0

    # n = 9 and 1 true bit
    >>> ising1d(np.array([0, 0, 0, 0, 0, 0, 0, 1, 0]))
    2

    # n = 9 and 1 true bit
    >>> ising1d(np.array([0, 0, 0, 0, 0, 1, 0, 0, 0]))
    2

    # n = 9 and 1 true bit
    >>> ising1d(np.array([0, 1, 0, 0, 0, 0, 0, 0, 0]))
    2

    # n = 9 and 2 true bits
    >>> ising1d(np.array([0, 1, 0, 1, 0, 0, 0, 0, 0]))
    4

    # n = 9 and 2 true bits
    >>> ising1d(np.array([0, 1, 0, 0, 1, 0, 0, 0, 0]))
    4

    # n = 9 and 2 true bits
    >>> ising1d(np.array([0, 1, 0, 1, 0, 0, 0, 0, 0]))
    4

    # n = 9 and 3 true bits
    >>> ising1d(np.array([0, 0, 1, 0, 0, 1, 0, 1, 0]))
    6

    # n = 9 and 3 true bits
    >>> ising1d(np.array([0, 0, 0, 1, 1, 0, 0, 1, 0]))
    4

    # n = 9 and 3 true bits
    >>> ising1d(np.array([0, 0, 0, 1, 0, 0, 0, 1, 1]))
    4

    # n = 9 and 4 true bits
    >>> ising1d(np.array([1, 0, 1, 1, 1, 0, 0, 0, 0]))
    4

    # n = 9 and 4 true bits
    >>> ising1d(np.array([1, 1, 1, 0, 1, 0, 0, 0, 0]))
    4

    # n = 9 and 4 true bits
    >>> ising1d(np.array([0, 1, 0, 0, 0, 1, 1, 0, 1]))
    6

    # n = 9 and 5 true bits
    >>> ising1d(np.array([0, 0, 1, 1, 0, 0, 1, 1, 1]))
    4

    # n = 9 and 5 true bits
    >>> ising1d(np.array([0, 1, 0, 1, 1, 0, 1, 0, 1]))
    8

    # n = 9 and 5 true bits
    >>> ising1d(np.array([1, 0, 1, 1, 1, 0, 0, 1, 0]))
    6

    # n = 9 and 6 true bits
    >>> ising1d(np.array([1, 0, 1, 1, 1, 1, 1, 0, 0]))
    4

    # n = 9 and 6 true bits
    >>> ising1d(np.array([0, 0, 1, 1, 1, 1, 1, 1, 0]))
    2

    # n = 9 and 6 true bits
    >>> ising1d(np.array([1, 1, 1, 1, 0, 1, 0, 0, 1]))
    4

    # n = 9 and 7 true bits
    >>> ising1d(np.array([1, 1, 0, 1, 1, 0, 1, 1, 1]))
    4

    # n = 9 and 7 true bits
    >>> ising1d(np.array([1, 1, 0, 1, 1, 1, 1, 1, 0]))
    4

    # n = 9 and 7 true bits
    >>> ising1d(np.array([1, 0, 1, 0, 1, 1, 1, 1, 1]))
    4

    # n = 9 and 8 true bits
    >>> ising1d(np.array([1, 0, 1, 1, 1, 1, 1, 1, 1]))
    2

    # n = 9 and 8 true bits
    >>> ising1d(np.array([1, 1, 1, 1, 1, 1, 0, 1, 1]))
    2

    # n = 9 and 8 true bits
    >>> ising1d(np.array([1, 1, 1, 1, 1, 0, 1, 1, 1]))
    2

    # n = 9 and 9 true bits
    >>> ising1d(np.array([1, 1, 1, 1, 1, 1, 1, 1, 1]))
    0

    # n = 10 and 0 true bits
    >>> ising1d(np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0]))
    0

    # n = 10 and 1 true bit
    >>> ising1d(np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 1]))
    2

    # n = 10 and 1 true bit
    >>> ising1d(np.array([0, 0, 0, 1, 0, 0, 0, 0, 0, 0]))
    2

    # n = 10 and 1 true bit
    >>> ising1d(np.array([1, 0, 0, 0, 0, 0, 0, 0, 0, 0]))
    2

    # n = 10 and 2 true bits
    >>> ising1d(np.array([0, 1, 0, 0, 0, 0, 0, 0, 1, 0]))
    4

    # n = 10 and 2 true bits
    >>> ising1d(np.array([1, 0, 0, 0, 0, 0, 0, 0, 0, 1]))
    2

    # n = 10 and 2 true bits
    >>> ising1d(np.array([0, 0, 0, 0, 1, 0, 0, 0, 0, 1]))
    4

    # n = 10 and 3 true bits
    >>> ising1d(np.array([1, 1, 1, 0, 0, 0, 0, 0, 0, 0]))
    2

    # n = 10 and 3 true bits
    >>> ising1d(np.array([0, 0, 1, 0, 0, 0, 1, 0, 1, 0]))
    6

    # n = 10 and 3 true bits
    >>> ising1d(np.array([0, 1, 0, 0, 0, 0, 1, 0, 0, 1]))
    6

    # n = 10 and 4 true bits
    >>> ising1d(np.array([1, 0, 0, 1, 1, 0, 0, 0, 1, 0]))
    6

    # n = 10 and 4 true bits
    >>> ising1d(np.array([0, 0, 0, 1, 1, 0, 1, 0, 0, 1]))
    6

    # n = 10 and 4 true bits
    >>> ising1d(np.array([1, 1, 0, 0, 0, 0, 0, 1, 0, 1]))
    4

    # n = 10 and 5 true bits
    >>> ising1d(np.array([1, 0, 0, 0, 1, 1, 1, 0, 1, 0]))
    6

    # n = 10 and 5 true bits
    >>> ising1d(np.array([1, 1, 0, 0, 1, 0, 1, 0, 1, 0]))
    8

    # n = 10 and 5 true bits
    >>> ising1d(np.array([0, 0, 1, 0, 0, 1, 0, 1, 1, 1]))
    6

    # n = 10 and 6 true bits
    >>> ising1d(np.array([1, 1, 1, 0, 0, 1, 1, 0, 0, 1]))
    4

    # n = 10 and 6 true bits
    >>> ising1d(np.array([1, 1, 0, 1, 0, 0, 1, 1, 1, 0]))
    6

    # n = 10 and 6 true bits
    >>> ising1d(np.array([1, 0, 0, 0, 1, 1, 1, 0, 1, 1]))
    4

    # n = 10 and 7 true bits
    >>> ising1d(np.array([1, 0, 0, 0, 1, 1, 1, 1, 1, 1]))
    2

    # n = 10 and 7 true bits
    >>> ising1d(np.array([1, 0, 0, 1, 1, 1, 1, 1, 1, 0]))
    4

    # n = 10 and 7 true bits
    >>> ising1d(np.array([0, 1, 1, 1, 0, 1, 0, 1, 1, 1]))
    6

    # n = 10 and 8 true bits
    >>> ising1d(np.array([1, 0, 1, 1, 0, 1, 1, 1, 1, 1]))
    4

    # n = 10 and 8 true bits
    >>> ising1d(np.array([1, 1, 1, 1, 1, 1, 1, 0, 1, 0]))
    4

    # n = 10 and 8 true bits
    >>> ising1d(np.array([0, 1, 1, 1, 1, 1, 1, 0, 1, 1]))
    4

    # n = 10 and 9 true bits
    >>> ising1d(np.array([1, 1, 1, 1, 1, 1, 1, 1, 0, 1]))
    2

    # n = 10 and 9 true bits
    >>> ising1d(np.array([1, 1, 1, 1, 1, 0, 1, 1, 1, 1]))
    2

    # n = 10 and 9 true bits
    >>> ising1d(np.array([1, 0, 1, 1, 1, 1, 1, 1, 1, 1]))
    2

    # n = 10 and 10 true bits
    >>> ising1d(np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1]))
    0
    """
    n: int = 0
    prev: bool = x[-1]
    for cur in x:
        if cur != prev:
            n += 1
        prev = cur
    return n


class Ising1d(BitStringProblem):
    """The one-dimensional Ising problem."""

    def __init__(self, n: int) -> None:
        """
        Initialize the one-dimensional Ising problem.

        :param n: the dimension of the problem

        >>> Ising1d(7).n
        7
        >>> Ising1d(3).evaluate(np.array([True, False, True]))
        2
        """
        super().__init__(n)
        self.evaluate = ising1d  # type: ignore

    def __str__(self) -> str:
        """
        Get the name of the one-dimensional Ising problem.

        :return: `ising1d_` + length of string

        >>> Ising1d(5)
        ising1d_5
        """
        return f"ising1d_{self.n}"

    @classmethod
    def default_instances(
            cls: type, scale_min: int = 2, scale_max: int = 100) \
            -> Iterator[Callable[[], "Ising1d"]]:
        """
        Get the 56 default instances of the :class:`Ising1d` problem.

        :param scale_min: the minimum permitted scale, by default `2`
        :param scale_max: the maximum permitted scale, by default `100`
        :returns: a sequence of default :class:`Ising1d` instances

        >>> len(list(Ising1d.default_instances()))
        56

        >>> [x() for x in Ising1d.default_instances()]
        [ising1d_2, ising1d_3, ising1d_4, ising1d_5, ising1d_6, ising1d_7, \
ising1d_8, ising1d_9, ising1d_10, ising1d_11, ising1d_12, ising1d_13, \
ising1d_14, ising1d_15, ising1d_16, ising1d_17, ising1d_18, ising1d_19, \
ising1d_20, ising1d_21, ising1d_22, ising1d_23, ising1d_24, ising1d_25, \
ising1d_26, ising1d_27, ising1d_28, ising1d_29, ising1d_30, ising1d_31, \
ising1d_32, ising1d_33, ising1d_36, ising1d_40, ising1d_41, ising1d_42, \
ising1d_44, ising1d_48, ising1d_49, ising1d_50, ising1d_55, ising1d_59, \
ising1d_60, ising1d_64, ising1d_66, ising1d_70, ising1d_77, ising1d_79, \
ising1d_80, ising1d_81, ising1d_85, ising1d_88, ising1d_90, ising1d_96, \
ising1d_99, ising1d_100]
        """
        return cast("Iterator[Callable[[], Ising1d]]",
                    super().default_instances(  # type: ignore
                        scale_min, scale_max))



"""
Example for bit-string based objective functions.

The problems noted here are mainly well-known benchmark problems from the
discrete optimization community. For these problems, we designed the dedicated
base class :class:`~moptipy.examples.bitstrings.bitstring_problem.\
BitStringProblem`.

The following benchmark problems are provided:

1.  The :mod:`~moptipy.examples.bitstrings.onemax` problem, where the
    goal is to find a bit string with the maximum number of ones.
2.  The :mod:`~moptipy.examples.bitstrings.leadingones` problem,
    where the goal is to find a bit string with the maximum number of leading
    ones.
3.  The :mod:`~moptipy.examples.bitstrings.linearharmonic` problem,
    where the goal is to find a bit string with the all ones, like in
    :mod:`~moptipy.examples.bitstrings.onemax`, but this time all bits have
    a different weight (namely their index, starting at 1).
4.  The :mod:`~moptipy.examples.bitstrings.binint` problem,
    is again similar to :mod:`~moptipy.examples.bitstrings.onemax`, but the
    bits have exponential weight. Basically, we just decode the bit string
    into an integer.
5.  The :mod:`~moptipy.examples.bitstrings.trap` problem, which is like
    OneMax, but with the optimum and worst-possible solution swapped. This
    problem is therefore highly deceptive.
6.  The :mod:`~moptipy.examples.bitstrings.twomax` problem has the global
    optimum at the string of all `1` bits and a local optimum at the string
    of all `0` bits. Both have basins of attraction of about the same size.
7.  :mod:`~moptipy.examples.bitstrings.ising1d`, the one-dimensional
    Ising model, where the goal is that all bits should have the same value as
    their neighbors in a ring.
8.  :mod:`~moptipy.examples.bitstrings.ising2d`, the two-dimensional
    Ising model, where the goal is that all bits should have the same value as
    their neighbors on a torus.
9.  The :mod:`~moptipy.examples.bitstrings.jump` problem is equivalent
    to :mod:`~moptipy.examples.bitstrings.onemax`, but has a deceptive
    region right before the optimum.
10. The :mod:`~moptipy.examples.bitstrings.plateau` problem similar to the
    :mod:`~moptipy.examples.bitstrings.jump` problem, but this time the
    optimum is surrounded by a region of neutrality.
11. The :mod:`~moptipy.examples.bitstrings.nqueens`, where the goal is to
    place `k` queens on a `k * k`-sized chess board such that no queen can
    beat any other queen.
12. The :mod:`~moptipy.examples.bitstrings.labs`, where the goal is to
    find a sequence with low autocorrelation and, thus, high merit factor.
    This problem is different from the others, because here, only for low
    values of `n`, the optimal solutions are actually known.
13. The :mod:`~moptipy.examples.bitstrings.w_model`, a benchmark
    problem with tunable epistasis, uniform neutrality, and
    ruggedness/deceptiveness.
14. The :mod:`~moptipy.examples.bitstrings.zeromax` problem, where the
    goal is to find a bit string with the maximum number of zeros. This is the
    opposite of the OneMax problem.

Parts of the code here are related to the research work of
Mr. Jiazheng ZENG (曾嘉政), a Master's student at the Institute of Applied
Optimization (应用优化研究所) of the School of
Artificial Intelligence and Big Data (人工智能与大数据学院) at
Hefei University (合肥大学) in
Hefei, Anhui, China (中国安徽省合肥市) under the supervision of
Prof. Dr. Thomas Weise (汤卫思教授).
"""
from itertools import chain
from typing import Callable, Iterable, Iterator, cast

from moptipy.examples.bitstrings.binint import BinInt
from moptipy.examples.bitstrings.bitstring_problem import BitStringProblem
from moptipy.examples.bitstrings.ising1d import Ising1d
from moptipy.examples.bitstrings.ising2d import Ising2d
from moptipy.examples.bitstrings.jump import Jump
from moptipy.examples.bitstrings.labs import LABS
from moptipy.examples.bitstrings.leadingones import LeadingOnes
from moptipy.examples.bitstrings.linearharmonic import LinearHarmonic
from moptipy.examples.bitstrings.nqueens import NQueens
from moptipy.examples.bitstrings.onemax import OneMax
from moptipy.examples.bitstrings.plateau import Plateau
from moptipy.examples.bitstrings.trap import Trap
from moptipy.examples.bitstrings.twomax import TwoMax
from moptipy.examples.bitstrings.w_model import WModel


def default_instances(  # noqa: RUF067
        class_scales: Callable[[
            type], Iterable[int]] = lambda _: cast("Iterable[int]", ())) \
        -> Iterator[Callable[[], BitStringProblem]]:
    """
    Get the default bit-string based benchmark instances.

    :param class_scales: a function that can override the minimum and
        maximimum problem scales on a per-benchmark-function-class
        basis. If this function returns an empty iterator, then the default
        scales are used.
    :return: an :class:`Iterable` with the default bit-string
        benchmark instances

    >>> len(list(default_instances()))
    1055

    >>> from moptipy.examples.bitstrings.binint import BinInt
    >>> from moptipy.examples.bitstrings.bitstring_problem import \
BitStringNKProblem
    >>> def clazz_scales(clazz) -> tuple[int, int]:
    ...     if issubclass(clazz, BinInt):
    ...         return 2, 30
    ...     if issubclass(clazz, BitStringNKProblem):
    ...         return 6, 16
    ...     return 8, 48
    >>> len(list(default_instances(clazz_scales)))
    331
    """
    return chain(
        BinInt.default_instances(*class_scales(BinInt)),  # type: ignore
        Ising1d.default_instances(*class_scales(Ising1d)),  # type: ignore
        Ising2d.default_instances(*class_scales(Ising2d)),  # type: ignore
        Jump.default_instances(*class_scales(Jump)),  # type: ignore
        LeadingOnes.default_instances(  # type: ignore
            *class_scales(LeadingOnes)),
        LinearHarmonic.default_instances(  # type: ignore
            *class_scales(LinearHarmonic)),
        NQueens.default_instances(*class_scales(NQueens)),  # type: ignore
        OneMax.default_instances(*class_scales(OneMax)),  # type: ignore
        Plateau.default_instances(*class_scales(Plateau)),  # type: ignore
        Trap.default_instances(*class_scales(Trap)),  # type: ignore
        TwoMax.default_instances(*class_scales(TwoMax)),  # type: ignore
        LABS.default_instances(*class_scales(LABS)),  # type: ignore
        WModel.default_instances(*class_scales(WModel)))  # type: ignore



"""
A temperature schedule as needed by Simulated Annealing.

The Simulated Annealing algorithm implemented in
:mod:`~moptipy.algorithms.so.simulated_annealing` performs a local search that
always accepts a non-worsening move, i.e., a solution which is not worse than
the currently maintained one. However, it will also *sometimes* accept one
that is worse. The probability of doing so depends on how much worse that
solution is and on the current *temperature* of the algorithm. The higher the
temperature, the higher the acceptance probability. The temperature changes
over time according to the
:class:`~moptipy.algorithms.modules.temperature_schedule.TemperatureSchedule`.

The temperature schedule receives an iteration index `tau` as input and
returns the current temperature via :meth:`~moptipy.algorithms.modules.\
temperature_schedule.TemperatureSchedule.temperature`. Notice that `tau` is
zero-based for simplicity reason, meanings that the first objective function
evaluation is at index `0`.
"""

from math import e, isfinite, log
from typing import Final

from pycommons.math.int_math import try_int
from pycommons.strings.enforce import enforce_non_empty_str_without_ws
from pycommons.types import check_int_range, type_error

from moptipy.api.component import Component
from moptipy.api.objective import Objective, check_objective
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.strings import num_to_str_for_name


# start schedule
class TemperatureSchedule(Component):
    """The base class for temperature schedules."""

    def __init__(self, t0: float) -> None:
        # end schedule
        """
        Initialize the temperature schedule.

        :param t0: the starting temperature, must be > 0
        """
        super().__init__()
        if not isinstance(t0, float):
            raise type_error(t0, "t0", float)
        if (not isfinite(t0)) or (t0 <= 0.0):
            raise ValueError(f"t0 must be >0, cannot be {t0}.")
# start schedule
        #: the starting temperature
        self.t0: Final[float] = t0

    def temperature(self, tau: int) -> float:
        """
        Compute the temperature at iteration `tau`.

        :param tau: the iteration index, starting with `0` at the first
            comparison of two solutions, at which point the starting
            temperature :attr:`~TemperatureSchedule.t0` should be returned
        :returns: the temperature
        """
# end schedule

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this temperature schedule as key-value pairs.

        :param logger: the logger for the parameters

        >>> from moptipy.utils.logger import InMemoryLogger
        >>> with InMemoryLogger() as l:
        ...     with l.key_values("C") as kv:
        ...         TemperatureSchedule(0.1).log_parameters_to(kv)
        ...     text = l.get_log()
        >>> text[1]
        'name: TemperatureSchedule'
        >>> text[3]
        'T0: 0.1'
        >>> len(text)
        6
        """
        super().log_parameters_to(logger)
        logger.key_value("T0", self.t0)


# start exponential
class ExponentialSchedule(TemperatureSchedule):
    """
    The exponential temperature schedule.

    The current temperature is computed as `t0 * (1 - epsilon) ** tau`.

    >>> ex = ExponentialSchedule(10.0, 0.05)
    >>> print(f"{ex.t0} - {ex.epsilon}")
    10.0 - 0.05
    >>> ex.temperature(0)
    10.0
    >>> ex.temperature(1)
    9.5
    >>> ex.temperature(2)
    9.025
    >>> ex.temperature(1_000_000_000_000_000_000)
    0.0
    """

    def __init__(self, t0: float, epsilon: float) -> None:
        """
        Initialize the exponential temperature schedule.

        :param t0: the starting temperature, must be > 0
        :param epsilon: the epsilon parameter of the schedule, in (0, 1)
        """
        super().__init__(t0)
# end exponential
        if not isinstance(epsilon, float):
            raise type_error(epsilon, "epsilon", float)
        if (not isfinite(epsilon)) or (not (0.0 < epsilon < 1.0)):
            raise ValueError(
                f"epsilon cannot be {epsilon}, must be in (0,1).")
# start exponential
        #: the epsilon parameter of the exponential schedule
        self.epsilon: Final[float] = epsilon
        #: the value used as basis for the exponent
        self.__one_minus_epsilon: Final[float] = 1.0 - epsilon
# end exponential
        if not (0.0 < self.__one_minus_epsilon < 1.0):
            raise ValueError(
                f"epsilon cannot be {epsilon}, because 1-epsilon must be in "
                f"(0, 1) but is {self.__one_minus_epsilon}.")
# start exponential

    def temperature(self, tau: int) -> float:
        """
        Compute the temperature at iteration `tau`.

        :param tau: the iteration index, starting with `0` at the first
            comparison of two solutions, at which point the starting
            temperature :attr:`~TemperatureSchedule.t0` should be returned
        :returns: the temperature

        >>> s = ExponentialSchedule(100.0, 0.5)
        >>> s.temperature(0)
        100.0
        >>> s.temperature(1)
        50.0
        >>> s.temperature(10)
        0.09765625
        """
        return self.t0 * (self.__one_minus_epsilon ** tau)
# end exponential

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of the exponential temperature schedule.

        :param logger: the logger for the parameters

        >>> from moptipy.utils.logger import InMemoryLogger
        >>> with InMemoryLogger() as l:
        ...     with l.key_values("C") as kv:
        ...         ExponentialSchedule(0.2, 0.6).log_parameters_to(kv)
        ...     text = l.get_log()
        >>> text[1]
        'name: exp0d2_0d6'
        >>> text[3]
        'T0: 0.2'
        >>> text[5]
        'e: 0.6'
        >>> len(text)
        8
        """
        super().log_parameters_to(logger)
        logger.key_value("e", self.epsilon)

    def __str__(self) -> str:
        """
        Get the string representation of the exponential temperature schedule.

        :returns: the name of this schedule

        >>> ExponentialSchedule(100.5, 0.3)
        exp100d5_0d3
        """
        return (f"exp{num_to_str_for_name(self.t0)}_"
                f"{num_to_str_for_name(self.epsilon)}")


# start logarithmic
class LogarithmicSchedule(TemperatureSchedule):
    """
    The logarithmic temperature schedule.

    The temperature is computed as `t0 / log(e + (tau * epsilon))`.

    >>> lg = LogarithmicSchedule(10.0, 0.1)
    >>> print(f"{lg.t0} - {lg.epsilon}")
    10.0 - 0.1
    >>> lg.temperature(0)
    10.0
    >>> lg.temperature(1)
    9.651322627630812
    >>> lg.temperature(1_000_000_000_000_000_000_000_000_000_000_000_000_000)
    0.11428802155348732
    """

    def __init__(self, t0: float, epsilon: float) -> None:
        """
        Initialize the logarithmic temperature schedule.

        :param t0: the starting temperature, must be > 0
        :param epsilon: the epsilon parameter of the schedule, is > 0
        """
        super().__init__(t0)
# end logarithmic
        if not isinstance(epsilon, float):
            raise type_error(epsilon, "epsilon", float)
        if (not isfinite(epsilon)) or (epsilon <= 0.0):
            raise ValueError(
                f"epsilon cannot be {epsilon}, must be > 0.")
# start logarithmic
        #: the epsilon parameter of the logarithmic schedule
        self.epsilon: Final[float] = epsilon

    def temperature(self, tau: int) -> float:
        """
        Compute the temperature at iteration `tau`.

        :param tau: the iteration index, starting with `0` at the first
            comparison of two solutions, at which point the starting
            temperature :attr:`~TemperatureSchedule.t0` should be returned
        :returns: the temperature

        >>> s = LogarithmicSchedule(100.0, 0.5)
        >>> s.temperature(0)
        100.0
        >>> s.temperature(1)
        85.55435113150568
        >>> s.temperature(10)
        48.93345190925178
        """
        return self.t0 / log(e + (tau * self.epsilon))
# end logarithmic

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of the logarithmic temperature schedule.

        :param logger: the logger for the parameters

        >>> from moptipy.utils.logger import InMemoryLogger
        >>> with InMemoryLogger() as l:
        ...     with l.key_values("C") as kv:
        ...         LogarithmicSchedule(0.2, 0.6).log_parameters_to(kv)
        ...     text = l.get_log()
        >>> text[1]
        'name: ln0d2_0d6'
        >>> text[3]
        'T0: 0.2'
        >>> text[5]
        'e: 0.6'
        >>> len(text)
        8
        """
        super().log_parameters_to(logger)
        logger.key_value("e", self.epsilon)

    def __str__(self) -> str:
        """
        Get the string representation of the logarithmic temperature schedule.

        :returns: the name of this schedule

        >>> LogarithmicSchedule(100.5, 0.3)
        ln100d5_0d3
        """
        return (f"ln{num_to_str_for_name(self.t0)}_"
                f"{num_to_str_for_name(self.epsilon)}")


class ExponentialScheduleBasedOnBounds(ExponentialSchedule):
    """
    An exponential schedule configured based on the objective's range.

    This exponential schedule takes an objective function as parameter.
    It uses the lower and the upper bound of this function, `LB` and `UB`,
    to select a start and end temperature based on the provided fractions.
    Here, we set `W = lb_sum_weight * LB + ub_sum_weight * UB`.
    If we set `lb_sum_weight = -1` and `ub_sum_weight = 1`, then `W` will be
    the range of the objective function.
    If we set `lb_sum_weight = 1` and `ub_sum_weight = 0`, then we base the
    temperature setup entirely on the lower bound.
    If we set `lb_sum_weight = 0` and `ub_sum_weight = 1`, then we base the
    temperature setup entirely on the upper bound.
    Roughly, the start temperature will be `W * start_range_frac` and
    the end temperature, to be reached after `n_steps` FEs, will be
    `W * end_range_frac`.
    Since sometimes the upper and lower bound may be excessivly large, we
    can provide limits for `W` in form of `min_bound_sum` and `max_bound_sum`.
    This will then override any other computation.
    Notice that it is expected that `tau == 0` when the temperature function
    is first called. It is expected that `tau == n_range - 1` when it is
    called for the last time.

    >>> from moptipy.examples.bitstrings.onemax import OneMax
    >>> es = ExponentialScheduleBasedOnBounds(
    ...     OneMax(10), -1, 1, 0.01, 0.0001, 10**8)
    >>> es.temperature(0)
    0.1
    >>> es.temperature(1)
    0.09999999539482989
    >>> es.temperature(10**8 - 1)
    0.0010000000029841878
    >>> es.temperature(10**8)
    0.0009999999569324865

    >>> es = ExponentialScheduleBasedOnBounds(
    ...         OneMax(10), -1, 1, 0.01, 0.0001, 10**8, max_bound_sum=5)
    >>> es.temperature(0)
    0.05
    >>> es.temperature(1)
    0.04999999769741494
    >>> es.temperature(10**8 - 1)
    0.0005000000014920939
    >>> es.temperature(10**8)
    0.0004999999784662432

    >>> try:
    ...     ExponentialScheduleBasedOnBounds(1, 0.01, 0.0001, 10**8)
    ... except TypeError as te:
    ...     print(te)
    objective function should be an instance of moptipy.api.objective.\
Objective but is int, namely 1.

    >>> try:
    ...     ExponentialScheduleBasedOnBounds(
    ...         OneMax(10), -1, 1, -1.0, 0.0001, 10**8)
    ... except ValueError as ve:
    ...     print(ve)
    Invalid bound sum factors [-1.0, 0.0001].

    >>> try:
    ...     ExponentialScheduleBasedOnBounds(
    ...         OneMax(10), -1, 1, 0.9, 0.0001, 1)
    ... except ValueError as ve:
    ...     print(ve)
    n_steps=1 is invalid, must be in 2..1000000000000000.
    """

    def __init__(self, f: Objective,
                 lb_sum_weight: int | float = -1,
                 ub_sum_weight: int | float = 1,
                 start_factor: float = 1e-3,
                 end_factor: float = 1e-7,
                 n_steps: int = 1_000_000,
                 min_bound_sum: int | float = 1e-20,
                 max_bound_sum: int | float = 1e20) -> None:
        """
        Initialize the range-based exponential schedule.

        :param f: the objective function whose range we will use
        :param lb_sum_weight: the weight of the lower bound in the bound sum
        :param ub_sum_weight: the weight of the upper bound in the bound sum
        :parma start_factor: the factor multiplied with the bound sum to get
            the starting temperature
        :parm end_factor: the factor multiplied with the bound sum to get
            the end temperature
        :param n_steps: the number of steps until the end range should be
            reached
        :param min_bound_sum: a lower limit for the weighted sum of the bounds
        :param max_bound_sum: an upper limit for the weighted sum of the bounds
        """
        f = check_objective(f)
        if not isinstance(lb_sum_weight, int | float):
            raise type_error(lb_sum_weight, "lb_sum_weight", (int, float))
        if not isinstance(ub_sum_weight, int | float):
            raise type_error(ub_sum_weight, "ub_sum_weight", (int, float))
        if not isinstance(start_factor, float):
            raise type_error(start_factor, "start_factor", float)
        if not isinstance(end_factor, float):
            raise type_error(end_factor, "end_factor", float)
        if not isinstance(min_bound_sum, int | float):
            raise type_error(min_bound_sum, "min_bound_sum", (int, float))
        if not isinstance(max_bound_sum, int | float):
            raise type_error(max_bound_sum, "max_bound_sum", (int, float))

        if not (isfinite(min_bound_sum) and isfinite(max_bound_sum) and (
                0 < min_bound_sum < max_bound_sum)):
            raise ValueError(f"Invalid bound sum limits [{min_bound_sum}"
                             f", {max_bound_sum}].")
        if not (isfinite(start_factor) and isfinite(end_factor) and (
                0.0 < end_factor < start_factor < 1e50)):
            raise ValueError(f"Invalid bound sum factors [{start_factor}"
                             f", {end_factor}].")
        if not (isfinite(lb_sum_weight) and isfinite(ub_sum_weight)):
            raise ValueError(f"Invalid bound sum weights [{lb_sum_weight}"
                             f", {ub_sum_weight}].")
        #: the number of steps that we will perform until reaching the end
        #: range fraction temperature
        self.__n_steps: Final[int] = check_int_range(
            n_steps, "n_steps", 2, 1_000_000_000_000_000)

        lb_sum_weight = try_int(lb_sum_weight)
        #: the sum weight for the lower bound
        self.__lb_sum_weight: Final[int | float] = lb_sum_weight
        ub_sum_weight = try_int(ub_sum_weight)
        #: the sum weight for the upper bound
        self.__ub_sum_weight: Final[int | float] = ub_sum_weight
        #: the start temperature bound sum factor
        self.__start_factor: Final[float] = start_factor
        #: the end temperature bound sum factor
        self.__end_factor: Final[float] = end_factor
        min_bound_sum = try_int(min_bound_sum)
        #: the minimum value for the bound sum
        self.__min_bound_sum: Final[int | float] = min_bound_sum
        max_bound_sum = try_int(max_bound_sum)
        #: the maximum value for the bound sum
        self.__max_bound_sum: Final[int | float] = max_bound_sum

        #: the name of the objective function used
        self.__used_objective: Final[str] = enforce_non_empty_str_without_ws(
            str(f))
        flb: Final[float | int] = f.lower_bound()
        fub: Final[float | int] = f.upper_bound()
        if flb > fub:
            raise ValueError(
                f"Objective function lower bound {flb} > upper bound {fub}?")

        #: the lower bound of the objective value
        self.__f_lower_bound: Final[int | float] = flb
        #: the upper bound for the objective value
        self.__f_upper_bound: Final[int | float] = fub

        bound_sum: Final[float | int] = try_int(max(min_bound_sum, min(
            max_bound_sum,
            (flb * lb_sum_weight if lb_sum_weight != 0 else 0) + (
                fub * ub_sum_weight if ub_sum_weight != 0 else 0))))
        if not (isfinite(bound_sum) and (
                min_bound_sum <= bound_sum <= max_bound_sum)):
            raise ValueError(
                f"Invalid bound sum {bound_sum} resulting from bounds [{flb}"
                f", {fub}] and weights {lb_sum_weight}, {ub_sum_weight}.")
        #: the bound sum
        self.__f_bound_sum: Final[int | float] = bound_sum

        t0: Final[float] = start_factor * bound_sum
        te: Final[float] = end_factor * bound_sum
        if not (isfinite(t0) and isfinite(te) and (0 < te < t0 < 1e100)):
            raise ValueError(
                f"Invalid setup {start_factor}, {end_factor}, "
                f"{bound_sum} leading to temperatures {t0}, {te}.")
        #: the end temperature
        self.__te: Final[float] = te

        epsilon: Final[float] = 1 - (te / t0) ** (1 / (n_steps - 1))
        if not (isfinite(epsilon) and (0 < epsilon < 1) and (
                0 < (1 - epsilon) < 1)):
            raise ValueError(
                f"Invalid computed epsilon {epsilon} resulting from setup "
                f"{start_factor}, {end_factor}, {bound_sum} leading "
                f"to temperatures {t0}, {te}.")
        super().__init__(t0, epsilon)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of the configured exponential temperature schedule.

        :param logger: the logger for the parameters

        >>> from moptipy.utils.logger import InMemoryLogger
        >>> from moptipy.examples.bitstrings.onemax import OneMax
        >>> with InMemoryLogger() as l:
        ...     with l.key_values("C") as kv:
        ...         ExponentialScheduleBasedOnBounds(
        ...             OneMax(10), -1, 1, 0.01, 0.0001).log_parameters_to(kv)
        ...     text = l.get_log()
        >>> text[1]
        'name: expRm1_1_0d01_0d0001_1em20_1e20'
        >>> text[3]
        'T0: 0.1'
        >>> text[5]
        'e: 4.6051641873212645e-6'
        >>> text[7]
        'nSteps: 1000000'
        >>> text[8]
        'lbSumWeight: -1'
        >>> text[9]
        'ubSumWeight: 1'
        >>> len(text)
        25
        """
        super().log_parameters_to(logger)
        logger.key_value("nSteps", self.__n_steps)
        logger.key_value("lbSumWeight", self.__lb_sum_weight)
        logger.key_value("ubSumWeight", self.__ub_sum_weight)
        logger.key_value("startFactor", self.__start_factor)
        logger.key_value("endFactor", self.__end_factor)
        logger.key_value("minBoundSum", self.__min_bound_sum)
        logger.key_value("maxBoundSum", self.__max_bound_sum)
        logger.key_value("f", self.__used_objective)
        logger.key_value("fLB", self.__f_lower_bound)
        logger.key_value("fUB", self.__f_upper_bound)
        logger.key_value("boundSum", self.__f_bound_sum)
        logger.key_value("Tend", self.__te)

    def __str__(self) -> str:
        """
        Get the string representation of the configured exponential schedule.

        :returns: the name of this schedule

        >>> from moptipy.examples.bitstrings.onemax import OneMax
        >>> ExponentialScheduleBasedOnBounds(OneMax(10), -1, 1, 0.01, 0.0001)
        expRm1_1_0d01_0d0001_1em20_1e20
        """
        return "expR" + "_".join(map(num_to_str_for_name, (
            self.__lb_sum_weight, self.__ub_sum_weight, self.__start_factor,
            self.__end_factor, self.__min_bound_sum, self.__max_bound_sum)))



"""
The BinInt problem maximizes the binary value of a bit string.

The BinInt problem is similar to the OneMax and LinearHarmonic in that it
tries to maximize the number of `True` bits in a string.
Different from these problems, however, it assigns exponentially increasing
weights to the bits.
The bit at index `1` has weight `2 ** (n - 1)`.
The bit at the last position has weight 1.
The upper bound for the objective function, reached when all bits are `False`,
therefore is `2 ** n - 1`. The lower bound, reached when all bits are `True`,
is `0`.

1. William Michael Rudnick. Genetic Algorithms and Fitness Variance with an
   Application to the Automated Design of Artiﬁcial Neural Networks.
   PhD thesis, Oregon Graduate Institute of Science & Technology: Beaverton,
   OR, USA, 1992. UMI Order No.: GAX92-22642.
2. Dirk Thierens, David Edward Goldberg, and Ângela Guimarães Pereira. Domino
   Convergence, Drift, and the Temporal-Salience Structure of Problems. In
   CEC'98, pages 535-540, 1998.
   doi: https://doi.org/10.1109/ICEC.1998.700085.
3. Kumara Sastry and David Edward Goldberg. Let's Get Ready to Rumble Redux:
   Crossover versus Mutation Head to Head on Exponentially Scaled Problems.
   In GECCO'07-I, pages 1380-1387, 2007.
   doi: https://doi.org10.1145/1276958.1277215.
4. Kumara Sastry and David Edward Goldberg. Let's Get Ready to Rumble Redux:
   Crossover versus Mutation Head to Head on Exponentially Scaled Problems.
   IlliGAL Report 2007005, Illinois Genetic Algorithms Laboratory (IlliGAL),
   Department of Computer Science, Department of General Engineering,
   University of Illinois at Urbana-Champaign: Urbana-Champaign, IL, USA,
   February 11, 2007.
"""

from typing import Callable, Final, Iterator, cast

import numba  # type: ignore
import numpy as np

from moptipy.examples.bitstrings.bitstring_problem import BitStringProblem


@numba.njit(nogil=True, cache=True)
def binint(x: np.ndarray) -> int:
    """
    Get the binint objective value: decode the inverted bit string as integer.

    :param x: the np array
    :return: the inverted bit string represented as integer

    >>> binint(np.array([0]))
    1
    >>> binint(np.array([1]))
    0

    >>> binint(np.array([0, 0]))
    3
    >>> binint(np.array([0, 1]))
    2
    >>> binint(np.array([1, 0]))
    1
    >>> binint(np.array([1, 1]))
    0

    >>> binint(np.array([0, 0, 0]))
    7
    >>> binint(np.array([0, 0, 1]))
    6
    >>> binint(np.array([0, 1, 0]))
    5
    >>> binint(np.array([0, 1, 1]))
    4
    >>> binint(np.array([1, 0, 0]))
    3
    >>> binint(np.array([1, 0, 1]))
    2
    >>> binint(np.array([1, 1, 0]))
    1
    >>> binint(np.array([1, 1, 1]))
    0
    """
    n: Final[int] = len(x)
    weight: int = 1 << n
    result: int = weight - 1
    for xx in x:
        weight >>= 1
        if xx:
            result -= weight
    return result


class BinInt(BitStringProblem):
    """Maximize the binary value of a bit string."""

    def __init__(self, n: int) -> None:
        """
        Initialize the binint objective function.

        :param n: the dimension of the problem

        >>> print(BinInt(2).n)
        2
        >>> print(BinInt(4).evaluate(np.array([True, True, False, True])))
        2
        """
        super().__init__(n)
        self.evaluate = binint  # type: ignore

    def __str__(self) -> str:
        """
        Get the name of the binint objective function.

        :return: `binint_` + length of string

        >>> BinInt(13)
        binint_13
        """
        return f"binint_{self.n}"

    def upper_bound(self) -> int:
        """
        Get the upper bound of the :class:`BinInt` problem.

        :returns: `(1 << n) - 1`

        >>> BinInt(4).upper_bound()
        15
        >>> BinInt(4).evaluate(np.array([0, 0, 0, 0]))
        15
        """
        return (1 << self.n) - 1

    @classmethod
    def default_instances(
            cls: type, scale_min: int = 2, scale_max: int = 30) \
            -> Iterator[Callable[[], "BinInt"]]:
        """
        Get the 29 default instances of the :class:`BinInt` problem.

        :param scale_min: the minimum permitted scale, by default `2`
        :param scale_max: the maximum permitted scale, by default `32`
        :returns: a sequence of default :class:`BinInt` instances

        >>> len(list(BinInt.default_instances()))
        29

        >>> [x() for x in BinInt.default_instances()]
        [binint_2, binint_3, binint_4, binint_5, binint_6, binint_7, \
binint_8, binint_9, binint_10, binint_11, binint_12, binint_13, binint_14, \
binint_15, binint_16, binint_17, binint_18, binint_19, binint_20, binint_21, \
binint_22, binint_23, binint_24, binint_25, binint_26, binint_27, binint_28, \
binint_29, binint_30]
        """
        return cast("Iterator[Callable[[], BinInt]]",
                    super().default_instances(  # type: ignore
                        scale_min, scale_max))



"""Test stuff on integer spaces."""

from typing import Any, Callable, Iterable, cast

import numpy as np
from numpy.random import Generator, default_rng
from pycommons.types import type_error

from moptipy.api.operators import Op0, Op1, Op2
from moptipy.spaces.intspace import IntSpace
from moptipy.tests.op0 import validate_op0
from moptipy.tests.op1 import validate_op1
from moptipy.tests.op2 import validate_op2


def intspaces_for_tests(
        space_filter: Callable[[IntSpace], bool] | None = None)\
        -> Iterable[IntSpace]:
    """
    Get a sequence of integer spaces for tests.

    :param perm_filter: an optional filter to sort out permutations we cannot
        use for testing
    :returns: the sequence of SignedPermutations
    """
    r = default_rng()
    pwrs: list[IntSpace] = [
        IntSpace(1, 0, 1),
        IntSpace(2, 0, 1),
        IntSpace(9, 0, 1),
        IntSpace(1, -1, 1),
        IntSpace(2, -1, 1),
        IntSpace(13, -1, 1),
        IntSpace(1, -12, 17),
        IntSpace(2, -11, 9),
        IntSpace(16, -23, 25),
        IntSpace(1, 0, 127),
        IntSpace(2, -128, 127),
        IntSpace(1, -128, 128),
        IntSpace(1, -2**15, 2**15 - 1),
        IntSpace(1, -2**15, 2**15),
        IntSpace(*map(int, (
            r.integers(1, 100), r.integers(0, 10), r.integers(10, 100)))),
        IntSpace(*map(int, (
            r.integers(1, 100), r.integers(-100, 0), r.integers(1, 100)))),
        IntSpace(*map(int, (
            r.integers(1, 100), r.integers(-100, -10),
            r.integers(-9, 100))))]
    if space_filter is not None:
        if not callable(space_filter):
            raise type_error(space_filter, "perm_filter", None, call=True)
        pwrs = [p for p in pwrs if space_filter(p)]
    r.shuffle(cast("list", pwrs))
    return pwrs


def make_ints_valid(space: IntSpace) -> \
        Callable[[Generator, np.ndarray], np.ndarray]:
    """
    Create a function that can make an int vector valid.

    :param space: the integer space
    :returns: the function
    """
    def __make_valid(prnd: Generator, x: np.ndarray, ppp=space) -> np.ndarray:
        np.copyto(x, prnd.integers(low=ppp.min_value, high=ppp.max_value,
                                   size=ppp.dimension, endpoint=True))
        return x
    return __make_valid


def validate_op0_on_1_intspace(
        op0: Op0 | Callable[[IntSpace], Op0],
        search_space: IntSpace,
        number_of_samples: int | None = None,
        min_unique_samples: int | Callable[[
            int, IntSpace], int] | None = None) -> None:
    """
    Validate the nullary operator on one `IntSpace` instance.

    :param op0: the operator or operator factory
    :param search_space: the search space
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    args: dict[str, Any] = {
        "op0": op0(search_space) if callable(op0) else op0,
        "search_space": search_space,
        "make_search_space_element_valid":
            make_ints_valid(search_space),
    }
    if number_of_samples is not None:
        args["number_of_samples"] = number_of_samples
    if min_unique_samples is not None:
        args["min_unique_samples"] = min_unique_samples
    validate_op0(**args)


def validate_op0_on_intspaces(
        op0: Op0 | Callable[[IntSpace], Op0],
        number_of_samples: int | None = None,
        min_unique_samples: int | Callable[[
            int, IntSpace], int] | None = None,
        space_filter: Callable[[IntSpace], bool] | None = None) \
        -> None:
    """
    Validate the nullary operator on several `IntSpace` instances.

    :param op0: the operator or operator factory
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    :param space_filter: an optional filter to sort out permutations we cannot
        use for testing
    """
    for pwr in intspaces_for_tests(space_filter):
        validate_op0_on_1_intspace(
            op0, pwr, number_of_samples, min_unique_samples)


def validate_op1_on_1_intspace(
        op1: Op1 | Callable[[IntSpace], Op1],
        search_space: IntSpace,
        number_of_samples: int | None = None,
        min_unique_samples: int | Callable[[
            int, IntSpace], int] | None = None) -> None:
    """
    Validate the unary operator on one `IntSpace` instance.

    :param op1: the operator or operator factory
    :param search_space: the search space
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    args: dict[str, Any] = {
        "op1": op1(search_space) if callable(op1) else op1,
        "search_space": search_space,
        "make_search_space_element_valid":
            make_ints_valid(search_space),
    }
    if number_of_samples is not None:
        args["number_of_samples"] = number_of_samples
    if min_unique_samples is not None:
        args["min_unique_samples"] = min_unique_samples
    validate_op1(**args)


def validate_op1_on_intspaces(
        op1: Op1 | Callable[[IntSpace], Op1],
        number_of_samples: int | None = None,
        min_unique_samples: int | Callable[[
            int, IntSpace], int] | None = None,
        space_filter: Callable[[IntSpace], bool] | None = None) \
        -> None:
    """
    Validate the unary operator on several `IntSpace` instances.

    :param op1: the operator or operator factory
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    :param space_filter: an optional filter to sort out permutations we cannot
        use for testing
    """
    for pwr in intspaces_for_tests(space_filter):
        validate_op1_on_1_intspace(
            op1, pwr, number_of_samples, min_unique_samples)


def validate_op2_on_1_intspace(
        op2: Op2 | Callable[[IntSpace], Op2],
        search_space: IntSpace,
        number_of_samples: int | None = None,
        min_unique_samples:
        int | Callable[[int, IntSpace], int] | None
        = None) -> None:
    """
    Validate the binary operator on one `IntSpace` instance.

    :param op2: the operator or operator factory
    :param search_space: the search space
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    args: dict[str, Any] = {
        "op2": op2(search_space) if callable(op2) else op2,
        "search_space": search_space,
        "make_search_space_element_valid": make_ints_valid(search_space),
    }
    if number_of_samples is not None:
        args["number_of_samples"] = number_of_samples
    if min_unique_samples is not None:
        args["min_unique_samples"] = min_unique_samples
    validate_op2(**args)


def validate_op2_on_intspaces(
        op2: Op2 | Callable[[IntSpace], Op2],
        number_of_samples: int | None = None,
        min_unique_samples:
        int | Callable[[int, IntSpace], int] | None = None,
        space_filter: Callable[[IntSpace], bool] | None = None) -> None:
    """
    Validate the binary operator on several `IntSpace` instances.

    :param op2: the operator or operator factory
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    :param space_filter: an optional filter to sort out permutations we cannot
        use for testing
    """
    for bst in intspaces_for_tests(space_filter):
        validate_op2_on_1_intspace(
            op2, bst, number_of_samples, min_unique_samples)



"""An implementation of a bit string based search space."""
from typing import Final

import numpy as np
from pycommons.io.csv import CSV_SEPARATOR
from pycommons.strings.string_conv import bool_to_str, str_to_bool
from pycommons.types import type_error

from moptipy.spaces.nparrayspace import NPArraySpace
from moptipy.utils.nputils import DEFAULT_BOOL


class BitStrings(NPArraySpace):
    """
    A space where each element is a bit string (:class:`numpy.ndarray`).

    With such a space, discrete optimization can be realized.

    >>> s = BitStrings(5)
    >>> print(s.dimension)
    5
    >>> print(s.dtype)
    bool
    >>> print(s.create())
    [False False False False False]
    >>> print(s.to_str(s.create()))
    FFFFF
    >>> print(s.from_str(s.to_str(s.create())))
    [False False False False False]
    """

    def __init__(self, dimension: int) -> None:
        """
        Create the vector-based search space.

        :param dimension: The dimension of the search space,
            i.e., the number of decision variables.
        """
        super().__init__(dimension, DEFAULT_BOOL)

    def create(self) -> np.ndarray:
        """
        Create a bit string filled with `False`.

        :return: the string

        >>> from moptipy.spaces.bitstrings import BitStrings
        >>> s = BitStrings(8)
        >>> v = s.create()
        >>> print(s.to_str(v))
        FFFFFFFF
        >>> print(v.dtype)
        bool
        """
        return np.zeros(shape=self.dimension, dtype=DEFAULT_BOOL)

    def from_str(self, text: str) -> np.ndarray:
        """
        Convert a string to a bit string.

        :param text: the text
        :return: the vector
        :raises TypeError: if `text` is not a `str`
        :raises ValueError: if `text` cannot be converted to a valid vector
        """
        if not (isinstance(text, str)):
            raise type_error(text, "text", str)
        x: Final[np.ndarray] = self.create()
        x[:] = [str_to_bool(t) for t in text]
        self.validate(x)
        return x

    def n_points(self) -> int:
        """
        Get the scale of the bit string space.

        :return: 2 ** dimension

        >>> print(BitStrings(4).n_points())
        16
        """
        return 1 << self.dimension  # = 2 ** self.dimension

    def __str__(self) -> str:
        """
        Get the name of this space.

        :return: "bits" + dimension

        >>> print(BitStrings(5))
        bits5
        """
        return f"bits{self.dimension}"

    def to_hex_string(self, x: np.ndarray) -> str:
        """
        Convert a bit string to a hexadecimal number.

        The reverse operator is given as :meth:`from_hex_string`.

        :param x: the bit string
        :returns: the hexadecimal number

        >>> b = BitStrings(134)
        >>> xt = b.create()
        >>> xt[:] = [((i % 3) == 0) != ((i % 7) == 0)
        ...         for i in range(b.dimension)]
        >>> "".join(reversed(b.to_str(xt)))
        'TTFFTFFFFFTFFTTFTFFTFTTFFTFFFFFTFFTTFTFFTFTTFFTFFFFFTFFTTFTFFTFT\
TFFTFFFFFTFFTTFTFFTFTTFFTFFFFFTFFTTFTFFTFTTFFTFFFFFTFFTTFTFFTFTTFFTFFF'
        >>> hex(int("11001000001001101001011001000001001101001011001000"
        ...         "0010011010010110010000010011010010110010000010011010"
        ...         "01011001000001001101001011001000", 2))
        '0x3209a5904d2c826964134b209a5904d2c8'
        >>> b.to_hex_string(xt)
        '3209a5904d2c826964134b209a5904d2c8'
        """
        self.validate(x)
        number: int = 0
        adder: int = 1
        for i in range(self.dimension):
            if x[i]:
                number += adder
            adder *= 2
        return format(number, "x")

    def from_hex_string(self, x: np.ndarray, s: str) -> None:
        """
        Convert a hexadecimal number to a numpy bit string.

        The reverse operator is given as :meth:`to_hex_string`.

        :param x: the bit string destination
        :param s: the string to convert

        >>> b = BitStrings(134)
        >>> x1 = b.create()
        >>> x1[:] = [((i % 3) == 0) != ((i % 7) == 0)
        ...         for i in range(b.dimension)]
        >>> x2 = b.create()
        >>> b.from_hex_string(x2, b.to_hex_string(x1))
        >>> b.is_equal(x1, x2)
        True
        """
        number: int = int(str.strip(s), 16)  # enforce type error
        for i in range(self.dimension):
            x[i] = (number & 1) != 0
            number >>= 1
        self.validate(x)

    def to_rle_str(self, x: np.ndarray) -> str:
        """
        Convert a numpy bit string to a run-length encoded string.

        The first bit value is stored, followed by a colon (:), followed by the
        run lengths separated with semicolons.
        The reverse operator is given as :meth:`from_rle_str`.

        :param x: the bit string
        :returns: the run-length encoded string

        >>> b = BitStrings(134)
        >>> xt = b.create()
        >>> xt[:] = [((i % 3) == 0) != ((i % 7) == 0)
        ...         for i in range(b.dimension)]
        >>> b.to_rle_str(xt)
        'F:3;1;2;2;1;1;2;1;1;2;2;1;5;1;2;2;1;1;2;1;1;2;2;1;5;1;2;2;1;1;2;1;1;\
2;2;1;5;1;2;2;1;1;2;1;1;2;2;1;5;1;2;2;1;1;2;1;1;2;2;1;5;1;2;2;1;1;2;1;1;2;2;\
1;5;1;2;2'

        >>> b = BitStrings(5)
        >>> xt = b.create()
        >>> xt.fill(0)
        >>> b.to_rle_str(xt)
        'F:5'
        >>> xt[1] = 1
        >>> b.to_rle_str(xt)
        'F:1;1;3'

        >>> b = BitStrings(5)
        >>> xt = b.create()
        >>> xt.fill(1)
        >>> b.to_rle_str(xt)
        'T:5'
        >>> xt[-1] = 0
        >>> b.to_rle_str(xt)
        'T:4;1'
        >>> xt[-1] = 1
        >>> xt[-2] = 0
        >>> b.to_rle_str(xt)
        'T:3;1;1'
        """
        self.validate(x)
        cur: np.bool = x[0]
        sep: str = ":"
        rle: list[str] = [bool_to_str(bool(cur))]
        start: int = 0
        for i, bit in enumerate(x):
            if bit != cur:
                rle.append(f"{sep}{i - start}")
                sep = CSV_SEPARATOR
                cur = bit
                start = i
        rle.extend(f"{sep}{self.dimension - start}")
        return "".join(rle)

    def from_rle_str(self, x: np.ndarray, s: str) -> None:
        """
        Convert a run-length encoded string to a numpy bit string.

        The reverse operator is given as :meth:`to_rle_str`.

        :param s: string to convert
        :param x: the bit string destination

        >>> b = BitStrings(134)
        >>> x1 = b.create()
        >>> x1[:] = [((i % 3) == 0) != ((i % 7) == 0)
        ...         for i in range(b.dimension)]
        >>> x2 = b.create()
        >>> b.from_hex_string(x2, b.to_hex_string(x1))
        >>> all(x1 == x2)
        True
        >>> x1[5] = not x1[5]
        >>> b.from_hex_string(x2, b.to_hex_string(x1))
        >>> all(x1 == x2)
        True

        >>> b = BitStrings(5)
        >>> x1 = b.create()
        >>> x2 = b.create()
        >>> x1.fill(0)
        >>> b.from_hex_string(x2, b.to_hex_string(x1))
        >>> all(x1 == x2)
        True

        >>> x1[1] = 1
        >>> b.from_hex_string(x2, b.to_hex_string(x1))
        >>> all(x1 == x2)
        True

        >>> x1.fill(1)
        >>> b.from_hex_string(x2, b.to_hex_string(x1))
        >>> all(x1 == x2)
        True
        """
        s = str.strip(s)
        value: bool = str_to_bool(s[0])
        i: int = 0
        for rl in s[s.index(":") + 1:].split(CSV_SEPARATOR):
            for _ in range(int(rl)):
                x[i] = value
                i += 1
            value = not value
        self.validate(x)
